# pylint: disable=missing-docstring
import io

import six

from django.apps import apps
from django.conf import settings
from django.contrib.auth import get_user_model
from django.contrib.auth.models import AnonymousUser, Group
from django.core.management import call_command
from django.test import override_settings

from guardian.shortcuts import assign_perm, remove_perm

# Import signals manually because we ignore them in App ready for tests
from resolwe.elastic import signals  # pylint: disable=unused-import
from resolwe.elastic.builder import index_builder
from resolwe.test import ElasticSearchTestCase

CUSTOM_SETTINGS = {
    'INSTALLED_APPS': settings.INSTALLED_APPS + ('resolwe.elastic.tests.test_app',),
}


@override_settings(**CUSTOM_SETTINGS)
class IndexTest(ElasticSearchTestCase):

    def setUp(self):
        apps.clear_cache()
        call_command('migrate', verbosity=0, interactive=False, load_initial_data=False)

        super().setUp()

    def test_mapping_multiple_times(self):
        index_builder.create_mappings()
        index_builder.create_mappings()

    def test_indexing(self):
        from .test_app.models import TestModel
        from .test_app.elastic_indexes import TestSearchDocument, TestSearchIndex

        # Create new object
        test_obj = TestModel.objects.create(name='Object name', number=43)

        es_objects = TestSearchDocument.search().execute()
        self.assertEqual(len(es_objects), 1)
        self.assertEqual(es_objects[0].name, 'Object name')
        self.assertEqual(es_objects[0].num, 43)
        self.assertEqual(es_objects[0].json['key'], 'value')

        # Update existing object
        test_obj.name = 'Another name'
        test_obj.save()

        es_objects = TestSearchDocument.search().execute()
        self.assertEqual(len(es_objects), 1)
        self.assertEqual(es_objects[0].name, 'Another name')
        self.assertEqual(es_objects[0].num, 43)
        self.assertEqual(es_objects[0].json['key'], 'value')

        # Create another object
        TestModel.objects.create(name='Another object', number=3)

        es_objects = TestSearchDocument.search().execute()
        self.assertEqual(len(es_objects), 2)

        # Delete object
        test_obj.delete()

        es_objects = TestSearchDocument.search().execute()
        self.assertEqual(len(es_objects), 1)

        # Create incorrect object (User object) and try to index it
        user_model = get_user_model()
        test_incorrect = user_model.objects.create(username='user_one')
        TestSearchIndex().build(test_incorrect)

    def test_bulk_indexing(self):
        from .test_app.models import TestModel
        from .test_app.elastic_indexes import TestSearchDocument

        first_obj = TestModel.objects.create(name='First name', number=42)
        TestModel.objects.create(name='Second name', number=43)

        # Delete whole index
        index_builder.delete()
        es_objects = TestSearchDocument.search().execute()
        self.assertEqual(len(es_objects), 0)

        # Build empty queryset
        index_builder.build(queryset=TestModel.objects.none())
        es_objects = TestSearchDocument.search().execute()
        self.assertEqual(len(es_objects), 0)

        # Build only the subset of queryset defined in index
        index_builder.build(queryset=TestModel.objects.filter(pk=first_obj.pk))
        es_objects = TestSearchDocument.search().execute()
        self.assertEqual(len(es_objects), 1)

        # Delete whole index
        index_builder.delete()
        es_objects = TestSearchDocument.search().execute()
        self.assertEqual(len(es_objects), 0)

        # Build only object
        index_builder.build(obj=first_obj)
        es_objects = TestSearchDocument.search().execute()
        self.assertEqual(len(es_objects), 1)

        # Delete whole index
        index_builder.delete()
        es_objects = TestSearchDocument.search().execute()
        self.assertEqual(len(es_objects), 0)

        # Build whole queryset defined in index
        index_builder.build()
        es_objects = TestSearchDocument.search().execute()
        self.assertEqual(len(es_objects), 2)

    def test_management_commands(self):
        from .test_app.models import TestModel
        from .test_app.elastic_indexes import TestSearchDocument, TestAnalyzerSearchDocument

        # Prepare test data
        TestModel.objects.create(name='Object name', number=43)

        es_objects = TestSearchDocument.search().execute()
        self.assertEqual(len(es_objects), 1)

        # Purge index
        call_command('elastic_purge', interactive=False, verbosity=0)

        es_objects = TestSearchDocument.search().execute()
        self.assertEqual(len(es_objects), 0)

        # Recreate index
        call_command('elastic_index', interactive=False, verbosity=0)

        es_objects = TestSearchDocument.search().execute()
        self.assertEqual(len(es_objects), 1)

        # Purge index
        call_command('elastic_purge', interactive=False, verbosity=0)

        es_objects = TestSearchDocument.search().execute()
        self.assertEqual(len(es_objects), 0)

        # Recreate only a specific index
        call_command('elastic_index', index=['TestAnalyzerSearchIndex'], interactive=False, verbosity=0)

        es_objects = TestSearchDocument.search().execute()
        self.assertEqual(len(es_objects), 0)
        es_objects = TestAnalyzerSearchDocument.search().execute()
        self.assertEqual(len(es_objects), 1)

        # Purge only a specific index
        call_command('elastic_purge', index=['TestAnalyzerSearchIndex'], interactive=False, verbosity=0)

        es_objects = TestAnalyzerSearchDocument.search().execute()
        self.assertEqual(len(es_objects), 0)

        call_command('elastic_index', exclude=['TestAnalyzerSearchIndex'], interactive=False, verbosity=0)
        es_objects = TestAnalyzerSearchDocument.search().execute()
        self.assertEqual(len(es_objects), 0)

        call_command('elastic_index', interactive=False, verbosity=0)
        es_objects = TestAnalyzerSearchDocument.search().execute()
        self.assertEqual(len(es_objects), 1)

        call_command('elastic_purge', exclude=['TestAnalyzerSearchIndex'], interactive=False, verbosity=0)

        es_objects = TestAnalyzerSearchDocument.search().execute()
        self.assertEqual(len(es_objects), 1)

        # Recreate an invalid index
        if six.PY2:
            output = io.BytesIO()
        else:
            output = io.StringIO()

        call_command('elastic_index', index=['InvalidIndex'], interactive=False, verbosity=0, stderr=output)
        self.assertIn("Unknown index: InvalidIndex", output.getvalue())

        call_command('elastic_index', exclude=['InvalidIndex'], interactive=False, verbosity=0, stderr=output)
        self.assertIn("Unknown index: InvalidIndex", output.getvalue())

        # Purge an invalid index
        call_command('elastic_purge', index=['InvalidIndex'], interactive=False, verbosity=0, stderr=output)
        self.assertIn("Unknown index: InvalidIndex", output.getvalue())

        call_command('elastic_purge', exclude=['InvalidIndex'], interactive=False, verbosity=0, stderr=output)
        self.assertIn("Unknown index: InvalidIndex", output.getvalue())

        # Create mappings.
        call_command('elastic_mapping', interactive=False, verbosity=0)

    def test_permissions(self):
        from .test_app.models import TestModel
        from .test_app.elastic_indexes import TestSearchDocument

        # Prepare users and groups
        user_model = get_user_model()
        user_1 = user_model.objects.create(username='user_one')
        user_2 = user_model.objects.create(username='user_two')
        user_3 = user_model.objects.create(username='user_three')
        user_pub = user_model.objects.get(username='public')
        group = Group.objects.create(name='group')

        # Create test object
        test_obj = TestModel.objects.create(name='Object name', number=43)
        assign_perm('view_testmodel', user_1, test_obj)
        assign_perm('view_testmodel', user_2, test_obj)
        assign_perm('view_testmodel', group, test_obj)

        es_objects = TestSearchDocument.search().execute()
        self.assertEqual(es_objects[0].users_with_permissions, [user_1.pk, user_2.pk])
        self.assertEqual(es_objects[0].groups_with_permissions, [group.pk])
        self.assertEqual(es_objects[0].public_permission, False)

        # Add user permission
        assign_perm('view_testmodel', user_3, test_obj)

        es_objects = TestSearchDocument.search().execute()
        self.assertEqual(es_objects[0].users_with_permissions, [user_1.pk, user_2.pk, user_3.pk])
        self.assertEqual(es_objects[0].groups_with_permissions, [group.pk])
        self.assertEqual(es_objects[0].public_permission, False)

        # Remove user permission
        remove_perm('view_testmodel', user_2, test_obj)

        es_objects = TestSearchDocument.search().execute()
        self.assertEqual(es_objects[0].users_with_permissions, [user_1.pk, user_3.pk])
        self.assertEqual(es_objects[0].groups_with_permissions, [group.pk])
        self.assertEqual(es_objects[0].public_permission, False)

        # Remove group permission
        remove_perm('view_testmodel', group, test_obj)

        es_objects = TestSearchDocument.search().execute()
        self.assertEqual(es_objects[0].users_with_permissions, [user_1.pk, user_3.pk])
        self.assertEqual(es_objects[0].groups_with_permissions, [])
        self.assertEqual(es_objects[0].public_permission, False)

        # Add group permission
        assign_perm('view_testmodel', group, test_obj)

        es_objects = TestSearchDocument.search().execute()
        self.assertEqual(es_objects[0].users_with_permissions, [user_1.pk, user_3.pk])
        self.assertEqual(es_objects[0].groups_with_permissions, [group.pk])
        self.assertEqual(es_objects[0].public_permission, False)

        # Add public permission
        assign_perm('view_testmodel', AnonymousUser(), test_obj)

        es_objects = TestSearchDocument.search().execute()
        self.assertEqual(es_objects[0].users_with_permissions, [user_pub.pk, user_1.pk, user_3.pk])
        self.assertEqual(es_objects[0].groups_with_permissions, [group.pk])
        self.assertEqual(es_objects[0].public_permission, True)

        # Remove public permission
        remove_perm('view_testmodel', AnonymousUser(), test_obj)

        es_objects = TestSearchDocument.search().execute()
        self.assertEqual(es_objects[0].users_with_permissions, [user_1.pk, user_3.pk])
        self.assertEqual(es_objects[0].groups_with_permissions, [group.pk])
        self.assertEqual(es_objects[0].public_permission, False)

    def test_field_name(self):
        from .test_app.models import TestModel
        from .test_app.elastic_indexes import TestSearchDocument

        TestModel.objects.create(name='Hello world FOO_BAR-G17-SA', number=42)

        es_objects = TestSearchDocument.search().execute()
        self.assertEqual(len(es_objects), 1)

        es_objects = TestSearchDocument.search().query('match', field_name='hello').execute()
        self.assertEqual(len(es_objects), 1)

        es_objects = TestSearchDocument.search().query('match', field_name='world').execute()
        self.assertEqual(len(es_objects), 1)

        es_objects = TestSearchDocument.search().query('match', **{'field_name.raw': 'hello'}).execute()
        self.assertEqual(len(es_objects), 0)

        es_objects = TestSearchDocument.search().query(
            'match', **{'field_name.raw': 'Hello world FOO_BAR-G17-SA'}
        ).execute()
        self.assertEqual(len(es_objects), 1)

        es_objects = TestSearchDocument.search().query('match', field_name='foo').execute()
        self.assertEqual(len(es_objects), 1)

        es_objects = TestSearchDocument.search().query('match', field_name='bar').execute()
        self.assertEqual(len(es_objects), 1)

        es_objects = TestSearchDocument.search().query('match', field_name='g17').execute()
        self.assertEqual(len(es_objects), 1)

        es_objects = TestSearchDocument.search().query('match', field_name='g17-sa').execute()
        self.assertEqual(len(es_objects), 1)

        es_objects = TestSearchDocument.search().query('match', field_name='17').execute()
        self.assertEqual(len(es_objects), 1)

    def test_field_process_type(self):
        from .test_app.models import TestModel
        from .test_app.elastic_indexes import TestSearchDocument

        TestModel.objects.create(field_process_type='data:geneset', number=42)
        TestModel.objects.create(field_process_type='data:geneset:venn', number=42)
        TestModel.objects.create(field_process_type='data:geneset:venn:omg', number=42)

        es_objects = TestSearchDocument.search().execute()
        self.assertEqual(len(es_objects), 3)

        es_objects = TestSearchDocument.search().query('match', field_process_type='data').execute()
        self.assertEqual(len(es_objects), 3)

        es_objects = TestSearchDocument.search().query('match', field_process_type='data:geneset').execute()
        self.assertEqual(len(es_objects), 3)

        es_objects = TestSearchDocument.search().query('match', field_process_type='data:geneset:venn').execute()
        self.assertEqual(len(es_objects), 2)

        es_objects = TestSearchDocument.search().query('match', field_process_type='data:geneset:venn:omg').execute()
        self.assertEqual(len(es_objects), 1)

        # Check if tokenizer did not include intermediate terms.
        es_objects = TestSearchDocument.search().query('match', field_process_type='geneset').execute()
        self.assertEqual(len(es_objects), 0)
        es_objects = TestSearchDocument.search().query('match', field_process_type='venn').execute()
        self.assertEqual(len(es_objects), 0)
        es_objects = TestSearchDocument.search().query('match', field_process_type='omg').execute()
        self.assertEqual(len(es_objects), 0)

    def test_dependencies(self):
        from .test_app.models import TestModelWithDependency, TestDependency
        from .test_app.elastic_indexes import TestModelWithDependencyDocument, TestModelWithFilterDependencyDocument

        model = TestModelWithDependency.objects.create(name='Deps')
        dep1 = TestDependency.objects.create(name='one')
        dep2 = TestDependency.objects.create(name='two')
        dep3 = TestDependency.objects.create(name='three')
        model.dependencies.add(dep1)
        model.dependencies.add(dep2)
        dep3.testmodelwithdependency_set.add(model)

        es_objects = TestModelWithDependencyDocument.search().query('match', name='deps').execute()
        self.assertEqual(len(es_objects), 1)

        es_objects = TestModelWithDependencyDocument.search().query('match', name='one').execute()
        self.assertEqual(len(es_objects), 1)

        es_objects = TestModelWithDependencyDocument.search().query('match', name='two').execute()
        self.assertEqual(len(es_objects), 1)

        es_objects = TestModelWithDependencyDocument.search().query('match', name='three').execute()
        self.assertEqual(len(es_objects), 1)

        es_objects = TestModelWithDependencyDocument.search().query('match', name='four').execute()
        self.assertEqual(len(es_objects), 0)

        dep3.name = 'four'
        dep3.save()

        es_objects = TestModelWithDependencyDocument.search().query('match', name='four').execute()
        self.assertEqual(len(es_objects), 1)

        es_objects = TestModelWithDependencyDocument.search().query('match', name='three').execute()
        self.assertEqual(len(es_objects), 0)

        dep3.delete()
        es_objects = TestModelWithDependencyDocument.search().query('match', name='four').execute()
        self.assertEqual(len(es_objects), 0)

        # Ensure that previous updates did not cause the filtered version to be updated.
        es_objects = TestModelWithFilterDependencyDocument.search().execute()
        self.assertEqual(len(es_objects), 1)
        # If the filtered version would be updated, this would instead equal 'Deps: one, two'.
        self.assertEqual(es_objects[0].name, 'Deps: ')

        dep4 = TestDependency.objects.create(name='hello')
        dep5 = TestDependency.objects.create(name='hello')
        model.dependencies.add(dep4)
        dep5.testmodelwithdependency_set.add(model)

        es_objects = TestModelWithFilterDependencyDocument.search().execute()
        self.assertEqual(len(es_objects), 1)
        # It is correct that even non-dependencies are contained in the name as dependencies are
        # only used to determine when to trigger updates.
        self.assertEqual(es_objects[0].name, 'Deps: one, two, hello, hello')

        model.dependencies.remove(dep4)
        dep5.testmodelwithdependency_set.remove(model)

        es_objects = TestModelWithFilterDependencyDocument.search().execute()
        self.assertEqual(len(es_objects), 1)
        # It is correct that even non-dependencies are contained in the name as dependencies are
        # only used to determine when to trigger updates.
        self.assertEqual(es_objects[0].name, 'Deps: one, two')

    def test_dependencies_reverse(self):
        from .test_app.models import TestModelWithDependency, TestDependency
        from .test_app.elastic_indexes import TestModelWithReverseDependencyDocument

        model1 = TestModelWithDependency.objects.create(name='One')
        model2 = TestModelWithDependency.objects.create(name='Two')
        model3 = TestModelWithDependency.objects.create(name='Three')
        dep = TestDependency.objects.create(name='deps')
        model1.dependencies.add(dep)
        model2.dependencies.add(dep)
        dep.testmodelwithdependency_set.add(model3)

        es_objects = TestModelWithReverseDependencyDocument.search().query('match', name='deps').execute()
        self.assertEqual(len(es_objects), 1)

        es_objects = TestModelWithReverseDependencyDocument.search().query('match', name='one').execute()
        self.assertEqual(len(es_objects), 1)

        es_objects = TestModelWithReverseDependencyDocument.search().query('match', name='two').execute()
        self.assertEqual(len(es_objects), 1)

        es_objects = TestModelWithReverseDependencyDocument.search().query('match', name='three').execute()
        self.assertEqual(len(es_objects), 1)

        es_objects = TestModelWithReverseDependencyDocument.search().query('match', name='four').execute()
        self.assertEqual(len(es_objects), 0)

        model3.name = 'Four'
        model3.save()

        es_objects = TestModelWithReverseDependencyDocument.search().query('match', name='four').execute()
        self.assertEqual(len(es_objects), 1)

        model3.delete()

        es_objects = TestModelWithReverseDependencyDocument.search().query('match', name='four').execute()
        self.assertEqual(len(es_objects), 0)

    def test_dependencies_self(self):
        from .test_app.models import TestSelfDependency
        from .test_app.elastic_indexes import TestModelWithSelfDependencyDocument

        dep1 = TestSelfDependency.objects.create(name='One')
        dep2 = TestSelfDependency.objects.create(name='Two')
        dep3 = TestSelfDependency.objects.create(name='Three')
        parent = TestSelfDependency.objects.create(name='Parent')
        parent.dependencies.add(dep1)
        parent.dependencies.add(dep2)
        dep3.parents.add(parent)

        es_objects = TestModelWithSelfDependencyDocument.search().query('match', name='parent').execute()
        self.assertEqual(len(es_objects), 1)
        self.assertEqual(es_objects[0].name, 'Parent: One, Two, Three')

        dep2.name = 'Too'
        dep2.save()

        es_objects = TestModelWithSelfDependencyDocument.search().query('match', name='parent').execute()
        self.assertEqual(len(es_objects), 1)
        self.assertEqual(es_objects[0].name, 'Parent: One, Too, Three')



# pylint: disable=missing-docstring
from __future__ import absolute_import, division, print_function, unicode_literals

from django.contrib.auth import get_user_model
from django.contrib.auth.models import AnonymousUser
from django.core.urlresolvers import reverse

from guardian.shortcuts import assign_perm
from rest_framework.test import APITestCase

from resolwe.flow.models import Process


class ProcessOrderingTest(APITestCase):

    def setUp(self):
        super(ProcessOrderingTest, self).setUp()

        user_model = get_user_model()
        user = user_model.objects.create(username='user')

        self.proc_1 = Process.objects.create(name='My process', contributor=user, version=1)
        self.proc_2 = Process.objects.create(name='My process', contributor=user, version=2)

        assign_perm('view_process', AnonymousUser(), self.proc_1)
        assign_perm('view_process', AnonymousUser(), self.proc_2)

        self.url = reverse('resolwe-api:process-list')

    def test_ordering_version(self):
        # pylint: disable=no-member
        response = self.client.get(self.url, {'ordering': 'version'}, format='json')
        self.assertEqual(response.data[0]['id'], self.proc_1.id)
        self.assertEqual(response.data[1]['id'], self.proc_2.id)

        response = self.client.get(self.url, {'ordering': '-version'}, format='json')
        self.assertEqual(response.data[0]['id'], self.proc_2.id)
        self.assertEqual(response.data[1]['id'], self.proc_1.id)



# Run Cellery with:
# DJANGO_SETTINGS_MODULE='tests.settings' celery worker -A tests --loglevel=info

from __future__ import absolute_import, division, print_function, unicode_literals

try:
    from celery import Celery
except ImportError:
    Celery = None

from django.conf import settings


app = None
if Celery:
    app = Celery('resolwe')

    app.config_from_object('django.conf:settings')
    app.autodiscover_tasks(lambda: settings.INSTALLED_APPS)



# -*- coding: utf-8 -*-
# Generated by Django 1.9.11 on 2016-11-30 16:09
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('flow', '0014_add_entity'),
    ]

    operations = [
        migrations.AlterField(
            model_name='collection',
            name='created',
            field=models.DateTimeField(auto_now_add=True, db_index=True),
        ),
        migrations.AlterField(
            model_name='collection',
            name='modified',
            field=models.DateTimeField(auto_now=True, db_index=True),
        ),
        migrations.AlterField(
            model_name='data',
            name='created',
            field=models.DateTimeField(auto_now_add=True, db_index=True),
        ),
        migrations.AlterField(
            model_name='data',
            name='finished',
            field=models.DateTimeField(blank=True, db_index=True, null=True),
        ),
        migrations.AlterField(
            model_name='data',
            name='modified',
            field=models.DateTimeField(auto_now=True, db_index=True),
        ),
        migrations.AlterField(
            model_name='data',
            name='started',
            field=models.DateTimeField(blank=True, db_index=True, null=True),
        ),
        migrations.AlterField(
            model_name='descriptorschema',
            name='created',
            field=models.DateTimeField(auto_now_add=True, db_index=True),
        ),
        migrations.AlterField(
            model_name='descriptorschema',
            name='modified',
            field=models.DateTimeField(auto_now=True, db_index=True),
        ),
        migrations.AlterField(
            model_name='entity',
            name='created',
            field=models.DateTimeField(auto_now_add=True, db_index=True),
        ),
        migrations.AlterField(
            model_name='entity',
            name='modified',
            field=models.DateTimeField(auto_now=True, db_index=True),
        ),
        migrations.AlterField(
            model_name='process',
            name='created',
            field=models.DateTimeField(auto_now_add=True, db_index=True),
        ),
        migrations.AlterField(
            model_name='process',
            name='modified',
            field=models.DateTimeField(auto_now=True, db_index=True),
        ),
        migrations.AlterField(
            model_name='storage',
            name='created',
            field=models.DateTimeField(auto_now_add=True, db_index=True),
        ),
        migrations.AlterField(
            model_name='storage',
            name='modified',
            field=models.DateTimeField(auto_now=True, db_index=True),
        ),
        migrations.AlterField(
            model_name='trigger',
            name='created',
            field=models.DateTimeField(auto_now_add=True, db_index=True),
        ),
        migrations.AlterField(
            model_name='trigger',
            name='modified',
            field=models.DateTimeField(auto_now=True, db_index=True),
        ),
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.9.6 on 2016-06-03 04:42
from __future__ import unicode_literals

import django.contrib.postgres.fields.jsonb
from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('flow', '0007_add_owner'),
    ]

    operations = [
        migrations.AlterField(
            model_name='collection',
            name='descriptor',
            field=django.contrib.postgres.fields.jsonb.JSONField(default=dict),
        ),
        migrations.AlterField(
            model_name='collection',
            name='settings',
            field=django.contrib.postgres.fields.jsonb.JSONField(default=dict),
        ),
        migrations.AlterField(
            model_name='data',
            name='descriptor',
            field=django.contrib.postgres.fields.jsonb.JSONField(default=dict),
        ),
        migrations.AlterField(
            model_name='data',
            name='input',
            field=django.contrib.postgres.fields.jsonb.JSONField(default=dict),
        ),
        migrations.AlterField(
            model_name='data',
            name='output',
            field=django.contrib.postgres.fields.jsonb.JSONField(default=dict),
        ),
        migrations.AlterField(
            model_name='descriptorschema',
            name='schema',
            field=django.contrib.postgres.fields.jsonb.JSONField(default=dict),
        ),
        migrations.AlterField(
            model_name='process',
            name='input_schema',
            field=django.contrib.postgres.fields.jsonb.JSONField(blank=True, default=list),
        ),
        migrations.AlterField(
            model_name='process',
            name='output_schema',
            field=django.contrib.postgres.fields.jsonb.JSONField(blank=True, default=list),
        ),
        migrations.AlterField(
            model_name='process',
            name='run',
            field=django.contrib.postgres.fields.jsonb.JSONField(default=dict),
        ),
        migrations.AlterField(
            model_name='trigger',
            name='input',
            field=django.contrib.postgres.fields.jsonb.JSONField(default=dict),
        ),
    ]



# pylint: disable=missing-docstring
from __future__ import absolute_import, division, print_function, unicode_literals

import mock

from resolwe.flow.execution_engines.bash import ExecutionEngine, ExecutionError
from resolwe.flow.expression_engines.jinja import ExpressionEngine
from resolwe.test import TestCase


class ExecutionEngineTestCase(TestCase):

    def test_invalid_template(self):
        manager = mock.MagicMock()
        manager.get_expression_engine.return_value = ExpressionEngine(manager)

        data_mock = mock.MagicMock(process_error=[])
        data_mock.process.requirements = {
            'expression-engine': 'jinja',
        }
        data_mock.process.run = {
            'language': 'bash',
            'program': '{% if reads.type.startswith("data:reads:") %}'
        }

        with self.assertRaises(ExecutionError):
            ExecutionEngine(manager).evaluate(data_mock)

        # If no expression engine is passed, then there should be no error.
        data_mock.process.requirements = {}
        result = ExecutionEngine(manager).evaluate(data_mock)
        self.assertEqual(result, data_mock.process.run['program'])


class CeleryEngineTestCase(TestCase):

    def test_passed_to_celery(self):
        pass



# pylint: disable=missing-docstring
from __future__ import absolute_import, division, print_function, unicode_literals

from copy import deepcopy

from django.contrib.auth import get_user_model
from django.contrib.auth.models import Group
from django.contrib.contenttypes.models import ContentType

from guardian.models import GroupObjectPermission, UserObjectPermission
from guardian.shortcuts import assign_perm, get_perms
from rest_framework import exceptions, status

from resolwe.flow.models import Collection, Data, DescriptorSchema, Entity, Process
from resolwe.flow.views import CollectionViewSet, DescriptorSchemaViewSet
from resolwe.permissions.utils import (
    assign_contributor_permissions, check_owner_permission, check_public_permissions, check_user_permissions,
)
from resolwe.test import ResolweAPITestCase, TestCase


class CollectionPermissionsTest(ResolweAPITestCase):

    def setUp(self):
        self.user1 = get_user_model().objects.create(username="test_user1")
        self.user2 = get_user_model().objects.create(username="test_user2")
        self.owner = get_user_model().objects.create(username="owner")
        # public user is already created bt django-guardian
        self.public = get_user_model().objects.get(username="public")

        self.group = Group.objects.create(name="Test group")

        self.collection = Collection.objects.create(
            contributor=self.user1,
            name="Test collection 1"
        )
        assign_perm('owner_collection', self.owner, self.collection)

        self.resource_name = 'collection'
        self.viewset = CollectionViewSet

        super(CollectionPermissionsTest, self).setUp()

    def _create_data(self):
        process = Process.objects.create(
            name='Test process',
            contributor=self.user1,
        )

        data = Data.objects.create(
            name='Test data',
            contributor=self.user1,
            process=process,
        )
        assign_perm('owner_data', self.owner, data)

        return data

    def _create_entity(self):
        return Entity.objects.create(
            name='Test entity',
            contributor=self.user1,
        )

    def test_public_user(self):
        """Public user cannot create/edit anything"""
        assign_perm("view_collection", self.user1, self.collection)
        assign_perm("share_collection", self.user1, self.collection)

        data = {'public': {'add': ['view']}}
        resp = self._detail_permissions(self.collection.pk, data, self.user1)
        self.assertEqual(resp.status_code, status.HTTP_200_OK)

        data = {'public': {'add': ['edit']}}
        resp = self._detail_permissions(self.collection.pk, data, self.user1)
        self.assertEqual(resp.status_code, status.HTTP_403_FORBIDDEN)

        data = {'public': {'remove': ['edit']}}
        resp = self._detail_permissions(self.collection.pk, data, self.user1)
        self.assertEqual(resp.status_code, status.HTTP_403_FORBIDDEN)

    def test_user_without_share(self):
        """User without ``SHARE`` permission cannot do anything"""
        assign_perm("view_collection", self.user1, self.collection)
        assign_perm("download_collection", self.user1, self.collection)
        assign_perm("add_collection", self.user1, self.collection)
        assign_perm("edit_collection", self.user1, self.collection)

        # Can not add permissions to users.
        data = {'users': {'add': {self.user2.pk: ['view']}}}
        resp = self._detail_permissions(self.collection.pk, data, self.user1)
        self.assertEqual(resp.status_code, status.HTTP_403_FORBIDDEN)

        # Can not add permissions to groups.
        data = {'users': {'add': {self.group.pk: ['view']}}}
        resp = self._detail_permissions(self.collection.pk, data, self.user1)
        self.assertEqual(resp.status_code, status.HTTP_403_FORBIDDEN)

        # Can not remove permissions from users.
        data = {'users': {'remove': {self.user2.pk: ['view']}}}
        resp = self._detail_permissions(self.collection.pk, data, self.user1)
        self.assertEqual(resp.status_code, status.HTTP_403_FORBIDDEN)

        # Can not remove permissions from groups.
        data = {'users': {'remove': {self.group.pk: ['view']}}}
        resp = self._detail_permissions(self.collection.pk, data, self.user1)
        self.assertEqual(resp.status_code, status.HTTP_403_FORBIDDEN)

    def test_user_with_share(self):
        assign_perm("view_collection", self.user1, self.collection)
        assign_perm("share_collection", self.user1, self.collection)

        # Can add permissions to users.
        data = {'users': {'add': {self.user2.pk: ['view']}}}
        resp = self._detail_permissions(self.collection.pk, data, self.user1)
        self.assertEqual(resp.status_code, status.HTTP_200_OK)
        self.assertEqual(UserObjectPermission.objects.filter(user=self.user2).count(), 1)

        # Can add permissions to groups.
        data = {'groups': {'add': {self.group.pk: ['view']}}}
        resp = self._detail_permissions(self.collection.pk, data, self.user1)
        self.assertEqual(resp.status_code, status.HTTP_200_OK)
        self.assertEqual(GroupObjectPermission.objects.count(), 1)

        # Can remove permissions from users.
        data = {'users': {'remove': {self.user2.pk: ['view']}}}
        resp = self._detail_permissions(self.collection.pk, data, self.user1)
        self.assertEqual(resp.status_code, status.HTTP_200_OK)
        self.assertEqual(UserObjectPermission.objects.filter(user=self.user2).count(), 0)

        # Can remove permissions from groups.
        data = {'groups': {'remove': {self.group.pk: ['view']}}}
        resp = self._detail_permissions(self.collection.pk, data, self.user1)
        self.assertEqual(resp.status_code, status.HTTP_200_OK)
        self.assertEqual(GroupObjectPermission.objects.count(), 0)

    def test_protect_owner(self):
        """Only owners can modify `owner` permission"""
        assign_perm("view_collection", self.user1, self.collection)
        assign_perm("download_collection", self.user1, self.collection)
        assign_perm("add_collection", self.user1, self.collection)
        assign_perm("edit_collection", self.user1, self.collection)
        assign_perm("share_collection", self.user1, self.collection)

        # User with share permission cannot grant ``owner`` permission
        data = {'users': {'add': {self.user2.pk: ['owner']}}}
        resp = self._detail_permissions(self.collection.pk, data, self.user1)
        self.assertEqual(resp.status_code, status.HTTP_403_FORBIDDEN)
        self.assertEqual(UserObjectPermission.objects.filter(user=self.user2).count(), 0)

        # User with share permission cannot revoke ``owner`` permission
        data = {'users': {'remove': {self.user2.pk: ['owner']}}}
        resp = self._detail_permissions(self.collection.pk, data, self.user1)
        self.assertEqual(resp.status_code, status.HTTP_403_FORBIDDEN)
        self.assertEqual(UserObjectPermission.objects.filter(user=self.user2).count(), 0)

        assign_perm("owner_collection", self.user1, self.collection)

        # ``owner`` permission cannot be assigned to a group
        data = {'groups': {'add': {self.group.pk: ['owner']}}}
        resp = self._detail_permissions(self.collection.pk, data, self.user1)
        self.assertEqual(resp.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertEqual(GroupObjectPermission.objects.filter(group=self.group).count(), 0)

        # User with owner permission can grant ``owner`` permission
        data = {'users': {'add': {self.user2.pk: ['owner']}}}
        resp = self._detail_permissions(self.collection.pk, data, self.user1)
        self.assertEqual(resp.status_code, status.HTTP_200_OK)
        self.assertEqual(UserObjectPermission.objects.filter(user=self.user2).count(), 1)

        # User with owner permission can revoke ``owner`` permission
        data = {'users': {'remove': {self.user2.pk: ['owner']}}}
        resp = self._detail_permissions(self.collection.pk, data, self.user1)
        self.assertEqual(resp.status_code, status.HTTP_200_OK)
        self.assertEqual(UserObjectPermission.objects.filter(user=self.user2).count(), 0)

        # User with owner permission cannot remove all owners
        data = {'users': {'remove': {self.user1.pk: ['owner'], self.owner.pk: ['owner']}}}
        resp = self._detail_permissions(self.collection.pk, data, self.user1)
        self.assertEqual(resp.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertEqual(resp.data['detail'], "Object must have at least one owner.")

        owner_count = UserObjectPermission.objects.filter(
            object_pk=self.collection.pk,
            content_type=ContentType.objects.get_for_model(self.collection),
            permission__codename__startswith='owner_'
        ).count()
        self.assertEqual(owner_count, 2)

        # User can delete his owner permission if there is at least one other owner
        data = {'users': {'remove': {self.user1.pk: ['owner']}}}
        resp = self._detail_permissions(self.collection.pk, data, self.user1)
        self.assertEqual(resp.status_code, status.HTTP_200_OK)

        user1_owner = UserObjectPermission.objects.filter(
            object_pk=self.collection.pk,
            content_type=ContentType.objects.get_for_model(self.collection),
            permission__codename__startswith='owner_',
            user=self.user1
        ).exists()
        self.assertFalse(user1_owner)

    def test_share_content(self):
        data_1, data_2 = self._create_data(), self._create_data()
        entity_1, entity_2 = self._create_entity(), self._create_entity()

        self.collection.data.add(data_1, data_2)
        self.collection.entity_set.add(entity_1, entity_2)

        assign_perm("view_collection", self.user1, self.collection)
        assign_perm("share_collection", self.user1, self.collection)
        assign_perm("view_data", self.user1, data_1)
        assign_perm("view_data", self.user1, data_2)
        assign_perm("share_data", self.user1, data_1)
        assign_perm("view_entity", self.user1, entity_1)
        assign_perm("view_entity", self.user1, entity_2)
        assign_perm("share_entity", self.user1, entity_1)

        data = {'users': {'add': {self.user2.pk: ['view']}}}
        resp = self._detail_permissions(self.collection.pk, data, self.user1)
        self.assertEqual(resp.status_code, status.HTTP_200_OK)
        self.assertEqual(UserObjectPermission.objects.filter(user=self.user2).count(), 1)

        data = {'users': {'add': {self.user2.pk: ['view']}}, 'share_content': 'true'}
        resp = self._detail_permissions(self.collection.pk, data, self.user1)
        self.assertEqual(resp.status_code, status.HTTP_200_OK)
        self.assertEqual(UserObjectPermission.objects.filter(user=self.user2).count(), 3)
        self.assertEqual(len(get_perms(self.user2, data_1)), 1)
        self.assertEqual(len(get_perms(self.user2, data_2)), 0)
        self.assertEqual(len(get_perms(self.user2, entity_1)), 1)
        self.assertEqual(len(get_perms(self.user2, entity_2)), 0)

        data = {'users': {'remove': {self.user2.pk: ['view']}}}
        resp = self._detail_permissions(self.collection.pk, data, self.user1)
        self.assertEqual(resp.status_code, status.HTTP_200_OK)
        self.assertEqual(UserObjectPermission.objects.filter(user=self.user2).count(), 2)

        data = {'users': {'remove': {self.user2.pk: ['view']}}, 'share_content': 'true'}
        resp = self._detail_permissions(self.collection.pk, data, self.user1)
        self.assertEqual(resp.status_code, status.HTTP_200_OK)
        self.assertEqual(UserObjectPermission.objects.filter(user=self.user2).count(), 0)

    def test_nonexisting_permission(self):
        assign_perm("view_collection", self.owner, self.collection)
        assign_perm("share_collection", self.owner, self.collection)

        # Add one valid permission to make sure that no permission is applied if any of them is unknown.
        data = {'users': {'add': {self.user1.pk: ['view'], self.user2.pk: ['view', 'foo']}}}
        resp = self._detail_permissions(self.collection.pk, data, self.owner)
        self.assertEqual(resp.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertEqual(resp.data['detail'], 'Unknown permission: foo')
        self.assertEqual(UserObjectPermission.objects.filter(user=self.user1).count(), 0)
        self.assertEqual(UserObjectPermission.objects.filter(user=self.user2).count(), 0)

        assign_perm("view_collection", self.user1, self.collection)
        assign_perm("view_collection", self.user2, self.collection)

        # Add one valid permission to make sure that no permission is applied if any of them is unknown.
        data = {'users': {'remove': {self.user1.pk: ['view'], self.user2.pk: ['view', 'foo']}}}
        resp = self._detail_permissions(self.collection.pk, data, self.owner)
        self.assertEqual(resp.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertEqual(resp.data['detail'], 'Unknown permission: foo')
        self.assertEqual(UserObjectPermission.objects.filter(user=self.user1).count(), 1)
        self.assertEqual(UserObjectPermission.objects.filter(user=self.user2).count(), 1)

    def test_nonexisting_user_group(self):
        assign_perm("view_collection", self.owner, self.collection)
        assign_perm("share_collection", self.owner, self.collection)

        user_perms_count = UserObjectPermission.objects.count()
        group_perms_count = GroupObjectPermission.objects.count()

        # Whole request should fail, so `user1` shouldn't have any permission assigned.
        data = {'users': {'add': {'999': ['view'], self.user1.pk: ['view']}}}
        resp = self._detail_permissions(self.collection.pk, data, self.owner)
        self.assertEqual(resp.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertEqual(resp.data['detail'], 'Unknown user: 999')
        self.assertEqual(UserObjectPermission.objects.count(), user_perms_count)

        # Whole request should fail, so `group` shouldn't have any permission assigned.
        data = {'groups': {'add': {'999': ['view'], self.group.pk: ['view']}}}
        resp = self._detail_permissions(self.collection.pk, data, self.owner)
        self.assertEqual(resp.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertEqual(resp.data['detail'], 'Unknown group: 999')
        self.assertEqual(GroupObjectPermission.objects.count(), group_perms_count)

        assign_perm("view_collection", self.user1, self.collection)
        assign_perm("view_collection", self.group, self.collection)

        user_perms_count = UserObjectPermission.objects.count()
        group_perms_count = GroupObjectPermission.objects.count()

        # Whole request should fail, so `user1` shouldn't have any permission removed.
        data = {'users': {'remove': {'999': ['view'], self.user1.pk: ['view']}}}
        resp = self._detail_permissions(self.collection.pk, data, self.owner)
        self.assertEqual(resp.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertEqual(resp.data['detail'], 'Unknown user: 999')
        self.assertEqual(UserObjectPermission.objects.count(), user_perms_count)

        # Whole request should fail, so `group` shouldn't have any permission removed.
        data = {'groups': {'remove': {'999': ['view'], self.group.pk: ['view']}}}
        resp = self._detail_permissions(self.collection.pk, data, self.owner)
        self.assertEqual(resp.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertEqual(resp.data['detail'], 'Unknown group: 999')
        self.assertEqual(GroupObjectPermission.objects.count(), group_perms_count)


class DescriptorSchemaPermissionsTest(ResolweAPITestCase):

    def setUp(self):
        self.user = get_user_model().objects.create(username='user')
        self.group = Group.objects.create(name='Test group')

        self.resource_name = 'collection'
        self.viewset = DescriptorSchemaViewSet

        super(DescriptorSchemaPermissionsTest, self).setUp()

        self.descriptor_schema = DescriptorSchema.objects.create(contributor=self.contributor)
        assign_contributor_permissions(self.descriptor_schema)

    def test_set_permissions(self):
        # Can add permissions to users.
        data = {'users': {'add': {self.user.pk: ['view']}}}
        resp = self._detail_permissions(self.descriptor_schema.pk, data, self.contributor)
        self.assertEqual(resp.status_code, status.HTTP_200_OK)
        self.assertEqual(UserObjectPermission.objects.filter(user=self.user).count(), 1)

        # Can add permissions to groups.
        data = {'groups': {'add': {self.group.pk: ['view']}}}
        resp = self._detail_permissions(self.descriptor_schema.pk, data, self.contributor)
        self.assertEqual(resp.status_code, status.HTTP_200_OK)
        self.assertEqual(GroupObjectPermission.objects.count(), 1)

        # Can remove permissions from users.
        data = {'users': {'remove': {self.user.pk: ['view']}}}
        resp = self._detail_permissions(self.descriptor_schema.pk, data, self.contributor)
        self.assertEqual(resp.status_code, status.HTTP_200_OK)
        self.assertEqual(UserObjectPermission.objects.filter(user=self.user).count(), 0)

        # Can remove permissions from groups.
        data = {'groups': {'remove': {self.group.pk: ['view']}}}
        resp = self._detail_permissions(self.descriptor_schema.pk, data, self.contributor)
        self.assertEqual(resp.status_code, status.HTTP_200_OK)
        self.assertEqual(GroupObjectPermission.objects.count(), 0)


class PermissionsUtilitiesTest(TestCase):

    def test_filter_owner_permission(self):
        """Check that ``owner`` permission is catched everywhere"""
        data_template = {
            'users': {
                'add': {
                    1: ['view'],
                    2: ['view', 'edit']
                },
                'remove': {
                    3: ['view', 'edit']
                }
            },
            'groups': {
                'add': {
                    1: ['view', 'edit']
                },
                'remove': {
                    2: ['view']
                }
            }
        }

        check_owner_permission(data_template, False)

        data = deepcopy(data_template)
        data['users']['add'][1].append('owner')
        with self.assertRaises(exceptions.PermissionDenied):
            check_owner_permission(data, False)
        check_owner_permission(data, True)

        data = deepcopy(data_template)
        data['users']['remove'][3].append('owner')
        with self.assertRaises(exceptions.PermissionDenied):
            check_owner_permission(data, False)
        check_owner_permission(data, True)

        data = deepcopy(data_template)
        data['groups']['add'][1].append('owner')
        with self.assertRaises(exceptions.ParseError):
            check_owner_permission(data, False)
        with self.assertRaises(exceptions.ParseError):
            check_owner_permission(data, True)

        data = deepcopy(data_template)
        data['groups']['remove'][2].append('owner')
        with self.assertRaises(exceptions.ParseError):
            check_owner_permission(data, False)
        with self.assertRaises(exceptions.ParseError):
            check_owner_permission(data, True)

    def test_filter_user_permissions(self):
        """Check that user cannot change his own permissions"""
        data = {
            'users': {
                'add': {
                    1: ['view'],
                },
                'remove': {
                    2: ['view', 'edit']
                }
            }
        }

        with self.assertRaises(exceptions.PermissionDenied):
            check_user_permissions(data, 1)

        with self.assertRaises(exceptions.PermissionDenied):
            check_user_permissions(data, 2)

        check_user_permissions(data, 3)

    def test_filter_public_permissions(self):
        """Check that public user cannot get to open permissions"""
        data = {'public': {'add': ['view']}}
        check_public_permissions(data)

        data = {'public': {'add': ['download']}}
        check_public_permissions(data)

        data = {'public': {'add': ['add']}}
        check_public_permissions(data)

        data = {'public': {'add': ['edit']}}
        with self.assertRaises(exceptions.PermissionDenied):
            check_public_permissions(data)

        data = {'public': {'add': ['share']}}
        with self.assertRaises(exceptions.PermissionDenied):
            check_public_permissions(data)

        data = {'public': {'add': ['owner']}}
        with self.assertRaises(exceptions.PermissionDenied):
            check_public_permissions(data)

        data = {'public': {'add': ['view', 'edit']}}
        with self.assertRaises(exceptions.PermissionDenied):
            check_public_permissions(data)



# -*- coding: utf-8 -*-
# Generated by Django 1.9 on 2016-03-08 06:56
from __future__ import unicode_literals

import django.contrib.postgres.fields.jsonb
from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('flow', '0002_project_to_collection'),
    ]

    operations = [
        migrations.AddField(
            model_name='collection',
            name='descriptor',
            field=django.contrib.postgres.fields.jsonb.JSONField(default={}),
        ),
        migrations.AddField(
            model_name='collection',
            name='descriptor_schema',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.PROTECT, to='flow.DescriptorSchema'),
        ),
        migrations.AddField(
            model_name='process',
            name='flow_collection',
            field=models.CharField(blank=True, max_length=100, null=True),
        ),
    ]



"""Custom database fields."""
from __future__ import absolute_import, division, print_function, unicode_literals

from six import string_types

from django.db import DatabaseError, connection
from django.db.models import constants
from django.db.models.fields import SlugField
from django.utils.text import slugify

MAX_SLUG_SEQUENCE_DIGITS = 9


class ResolweSlugField(SlugField):
    """Slug field."""

    def __init__(self, *args, **kwargs):
        """Initialize field."""
        self.populate_from = kwargs.pop('populate_from', None)

        # unique_with value can be string or tuple
        self.unique_with = kwargs.pop('unique_with', ())
        if isinstance(self.unique_with, string_types):
            self.unique_with = (self.unique_with,)

        if self.unique_with:
            # we will do "manual" granular check below
            kwargs['unique'] = False

        super(ResolweSlugField, self).__init__(*args, **kwargs)

    def deconstruct(self):
        """Deconstruct method."""
        name, path, args, kwargs = super(ResolweSlugField, self).deconstruct()

        if self.populate_from is not None:
            kwargs['populate_from'] = self.populate_from

        if self.unique_with != ():
            kwargs['unique_with'] = self.unique_with
            kwargs.pop('unique', None)

        return name, path, args, kwargs

    def _get_unique_constraints(self, instance):
        """Return SQL filter for filtering by fields in ``unique_with`` attribute.

        Filter is returned as tuple of two elements where first one is
        placeholder which is safe to insert into SQL query and second
        one may include potentially dangerous values and must be passed
        to SQL query in ``params`` attribute to make sure it is properly
        escaped.
        """
        constraints_expression = []
        constraints_values = {}
        for field_name in self.unique_with:
            if constants.LOOKUP_SEP in field_name:
                raise NotImplementedError(
                    '`unique_with` constraint does not support lookups by related models.'
                )

            field = instance._meta.get_field(field_name)  # pylint: disable=protected-access
            field_value = getattr(instance, field_name)

            # Convert value to the database representation.
            field_db_value = field.get_prep_value(field_value)

            constraint_key = 'unique_' + field_name
            constraints_expression.append("{} = %({})s".format(
                connection.ops.quote_name(field.column),
                constraint_key
            ))
            constraints_values[constraint_key] = field_db_value

        if not constraints_expression:
            return '', []

        constraints_expression = 'AND ' + ' AND '.join(constraints_expression)
        return constraints_expression, constraints_values

    def _get_populate_from_value(self, instance):
        """Get the value from ``populate_from`` attribute."""
        if hasattr(self.populate_from, '__call__'):
            # ResolweSlugField(populate_from=lambda instance: ...)
            return self.populate_from(instance)
        else:
            # ResolweSlugField(populate_from='foo')
            attr = getattr(instance, self.populate_from)
            return attr() if callable(attr) else attr

    def pre_save(self, instance, add):
        """Ensure slug uniqunes before save."""
        slug = self.value_from_object(instance)

        # We don't want to change slug defined by user.
        predefined_slug = bool(slug)

        if not slug and self.populate_from:
            slug = self._get_populate_from_value(instance)

        if slug:
            slug = slugify(slug)

        if not slug:
            slug = None

            if not self.blank:
                slug = instance._meta.model_name  # pylint: disable=protected-access
            elif not self.null:
                slug = ''

        if slug:
            # Make sure that auto generated slug with added sequence
            # won't excede maximal length.
            # Validation of predefined slugs is handled by Django.
            if not predefined_slug:
                slug = slug[:(self.max_length - MAX_SLUG_SEQUENCE_DIGITS - 1)]

            constraints_placeholder, constraints_values = self._get_unique_constraints(instance)

            instance_pk_name = instance._meta.pk.name  # pylint: disable=protected-access

            # Safe values - make sure that there is no chance of SQL injection.
            query_params = {
                'constraints_placeholder': constraints_placeholder,
                'slug_column': connection.ops.quote_name(self.column),
                'slug_len': len(slug),
                'table_name': connection.ops.quote_name(self.model._meta.db_table),  # pylint: disable=protected-access
                'pk_neq_placeholder': 'AND {} != %(instance_pk)s'.format(instance_pk_name) if instance.pk else ''
            }

            # SQL injection unsafe values - will be escaped.
            # Keys prefixed with `unique_` are reserved for `constraints_values` dict.
            query_escape_params = {
                'slug': slug,
                'slug_regex': '^{}(-[0-9]*)?$'.format(slug),
            }
            query_escape_params.update(constraints_values)
            if instance.pk:
                query_escape_params['instance_pk'] = instance.pk

            with connection.cursor() as cursor:
                # TODO: Slowest part of this query is `MAX` function. It can
                #       be optimized by indexing slug column by slug sequence.
                #       https://www.postgresql.org/docs/9.4/static/indexes-expressional.html
                cursor.execute(
                    """
                    SELECT
                        CASE
                            WHEN (
                                EXISTS(
                                    SELECT 1 FROM {table_name} WHERE (
                                        {slug_column} = %(slug)s
                                        {pk_neq_placeholder}
                                        {constraints_placeholder}
                                    )
                                )
                            ) THEN MAX(slug_sequence) + 1
                            ELSE NULL
                        END
                    FROM (
                        SELECT COALESCE(
                            NULLIF(
                                RIGHT({slug_column}, -{slug_len}-1),
                                ''
                            ),
                            '1'
                        )::text::integer AS slug_sequence
                        FROM {table_name} WHERE (
                            {slug_column} ~ %(slug_regex)s
                            {pk_neq_placeholder}
                            {constraints_placeholder}
                        )
                    ) AS tmp
                    """.format(**query_params),
                    params=query_escape_params
                )
                result = cursor.fetchone()[0]

            if result is not None:
                if predefined_slug:
                    raise DatabaseError(
                        "Slug '{}' (version {}) is already taken.".format(slug, instance.version)
                    )

                if len(str(result)) > MAX_SLUG_SEQUENCE_DIGITS:
                    raise DatabaseError(
                        "Auto-generated slug sequence too long - please choose a different slug."
                    )

                slug = '{}-{}'.format(slug, result)

            # Make the updated slug available as instance attribute.
            setattr(instance, self.name, slug)

        return slug



""".. Ignore pydocstyle D400.

====
Flow
====

"""
from __future__ import absolute_import, division, print_function, unicode_literals

default_app_config = 'resolwe.flow.apps.FlowConfig'  # pylint: disable=invalid-name



""".. Ignore pydocstyle D400.

.. autoclass:: resolwe.test.ResolweAPITestCase
    :members:

"""

from __future__ import absolute_import, division, print_function, unicode_literals

import six

from django.contrib.auth import get_user_model
from django.core.urlresolvers import reverse

from rest_framework.test import APIRequestFactory, APITestCase, force_authenticate

from resolwe.test import TestCaseHelpers


class ResolweAPITestCase(TestCaseHelpers, APITestCase):
    """Base class for testing Resolwe REST API.

    This class is derived from Django REST Framework's
    :drf:`APITestCase <testing/#test-cases>` class and has implemented
    some basic features that makes testing Resolwe API easier. These
    features includes following functions:

    .. automethod:: _get_list
    .. automethod:: _get_detail
    .. automethod:: _post
    .. automethod:: _patch
    .. automethod:: _delete
    .. automethod:: _detail_permissions

    It also has included 2 views made from referenced DRF's ``ViewSet``.
    First mimic list view and has following links between request's
    methods and ViewSet's methods:

      *  ``GET`` -> ``list``
      *  ``POST`` -> ``create``

    Second mimic detail view and has following links between request's
    methods and ViewSet's methods:

      *  ``GET`` -> ``retrieve``
      *  ``PUT`` -> ``update``
      *  ``PATCH`` -> ``partial_update``
      *  ``DELETE`` -> ``destroy``
      *  ``POST`` -> ``permissions``

    If any of the listed methods is not defined in the VievSet,
    corresponding link is omitted.

    .. note::
        ``self.viewset`` (instance of DRF's ``Viewset``) and
        ``self.resource_name`` (string) must be defined before calling
        super ``setUp`` method to work properly.

    ``self.factory`` is instance of DRF's ``APIRequestFactory``.

    """

    def setUp(self):
        """Prepare data."""
        super(ResolweAPITestCase, self).setUp()

        # TODO: Remove this when removing fixtures
        if get_user_model().objects.filter(pk=2).exists():
            self.user1 = get_user_model().objects.get(pk=2)
        if get_user_model().objects.filter(pk=3).exists():
            self.user2 = get_user_model().objects.get(pk=3)
        if get_user_model().objects.filter(pk=4).exists():
            self.user3 = get_user_model().objects.get(pk=4)
        if get_user_model().objects.filter(pk=5).exists():
            self.admin = get_user_model().objects.get(pk=5)

        user_model = get_user_model()
        # TODO: Change username to `admin` when fixtures are removed
        self.admin = user_model.objects.create_superuser(username='admin2', email='admin@test.com', password='admin')
        self.contributor = user_model.objects.create_user(username='contributor')

        if not hasattr(self, 'viewset'):
            raise KeyError("`self.viewset` must be defined in child class")

        if not hasattr(self, 'resource_name'):
            raise KeyError("`self.resource_name` must be defined in child class")

        self.factory = APIRequestFactory()

        list_url_mapping = {}
        if hasattr(self.viewset, 'list'):  # pylint: disable=no-member
            list_url_mapping['get'] = 'list'
        if hasattr(self.viewset, 'create'):  # pylint: disable=no-member
            list_url_mapping['post'] = 'create'

        self.list_view = self.viewset.as_view(list_url_mapping)  # pylint: disable=no-member

        detail_url_mapping = {}
        if hasattr(self.viewset, 'retrieve'):  # pylint: disable=no-member
            detail_url_mapping['get'] = 'retrieve'
        if hasattr(self.viewset, 'update'):  # pylint: disable=no-member
            detail_url_mapping['put'] = 'update'
        if hasattr(self.viewset, 'partial_update'):  # pylint: disable=no-member
            detail_url_mapping['patch'] = 'partial_update'
        if hasattr(self.viewset, 'destroy'):  # pylint: disable=no-member
            detail_url_mapping['delete'] = 'destroy'
        if hasattr(self.viewset, 'detail_permissions'):  # pylint: disable=no-member
            detail_url_mapping['post'] = 'detail_permissions'

        self.detail_view = self.viewset.as_view(detail_url_mapping)  # pylint: disable=no-member

    def detail_url(self, pk):
        """Get detail url."""
        return reverse('resolwe-api:{}-detail'.format(self.resource_name), kwargs={'pk': pk})  # noqa pylint: disable=no-member

    def detail_permissions(self, pk):
        """Get detail permissions url."""
        return reverse('resolwe-api:{}-permissions'.format(self.resource_name), kwargs={'pk': pk})  # noqa pylint: disable=no-member

    @property
    def list_url(self):
        """Get list url."""
        return reverse('resolwe-api:{}-list'.format(self.resource_name))  # pylint: disable=no-member

    def _render_query_params(self, params):
        """Generate query parameters from given dict."""
        if not params:
            return ''

        return '?' + '&'.join('{}={}'.format(key, value) for key, value in six.iteritems(params))

    def _get_list(self, user=None, query_params={}):
        """Make ``GET`` request to ``self.list_view`` view.

        If ``user`` is not ``None``, the given user is authenticated
        before making the request.

        :param user: User to authenticate in request
        :type user: :class:`~django.contrib.auth.models.User` or :data:`None`
        :return: API response object
        :rtype: :drf:`Response <responses/#response>`

        """
        url = self.list_url + self._render_query_params(query_params)
        request = self.factory.get(url, format='json')
        force_authenticate(request, user)
        return self.list_view(request)

    def _get_detail(self, pk, user=None, query_params={}):
        """Make ``GET`` request to ``self.detail_view`` view.

        If ``user`` is not ``None``, the given user is authenticated
        before making the request.

        :param int pk: Primary key of the coresponding object
        :param user: User to authenticate in request
        :type user: :class:`~django.contrib.auth.models.User` or :data:`None`
        :return: API response object
        :rtype: :drf:`Response <responses/#response>`

        """
        url = self.detail_url(pk) + self._render_query_params(query_params)
        request = self.factory.get(url, format='json')
        force_authenticate(request, user)
        return self.detail_view(request, pk=pk)

    def _post(self, data={}, user=None, query_params={}):
        """Make ``POST`` request to ``self.list_view`` view.

        If ``user`` is not ``None``, the given user is authenticated
        before making the request.

        :param dict data: data for posting in request's body
        :param user: User to authenticate in request
        :type user: :class:`~django.contrib.auth.models.User` or :data:`None`
        :return: API response object
        :rtype: :drf:`Response <responses/#response>`

        """
        url = self.list_url + self._render_query_params(query_params)
        request = self.factory.post(url, data=data, format='json')
        force_authenticate(request, user)
        return self.list_view(request)

    def _patch(self, pk, data={}, user=None, query_params={}):
        """Make ``PATCH`` request to ``self.detail_view`` view.

        If ``user`` is not ``None``, the given user is authenticated
        before making the request.

        :param int pk: Primary key of the coresponding object
        :param dict data: data for posting in request's body
        :param user: User to authenticate in request
        :type user: :class:`~django.contrib.auth.models.User` or :data:`None`
        :return: API response object
        :rtype: :drf:`Response <responses/#response>`

        """
        url = self.detail_url(pk) + self._render_query_params(query_params)
        request = self.factory.patch(url, data=data, format='json')
        force_authenticate(request, user)
        return self.detail_view(request, pk=pk)

    def _delete(self, pk, user=None, query_params={}):
        """Make ``DELETE`` request to ``self.detail_view`` view.

        If ``user`` is not ``None``, the given user is authenticated
        before making the request.

        :param int pk: Primary key of the coresponding object
        :param user: User to authenticate in request
        :type user: :class:`~django.contrib.auth.models.User` or :data:`None`
        :return: API response object
        :rtype: :drf:`Response <responses/#response>`

        """
        url = self.detail_url(pk) + self._render_query_params(query_params)
        request = self.factory.delete(url, format='json')
        force_authenticate(request, user)
        return self.detail_view(request, pk=pk)

    def _detail_permissions(self, pk, data={}, user=None):
        """Make ``POST`` request to ``self.detail_view`` view.

        If ``user`` is not ``None``, the given user is authenticated
        before making the request.

        :param int pk: Primary key of the coresponding object
        :param dict data: data for posting in request's body
        :param user: User to authenticate in request
        :type user: :class:`~django.contrib.auth.models.User` or :data:`None`
        :return: API response object
        :rtype: :drf:`Response <responses/#response>`

        """
        request = self.factory.post(self.detail_permissions(pk), data=data, format='json')
        force_authenticate(request, user)
        return self.detail_view(request, pk=pk)

    def assertKeys(self, data, wanted):  # pylint: disable=invalid-name
        """Assert dictionary keys."""
        self.assertEqual(sorted(data.keys()), sorted(wanted))



"""Support for selective field serialization."""
from __future__ import absolute_import, division, print_function, unicode_literals

import copy

from .projection import apply_subfield_projection


class SelectiveFieldMixin(object):
    """Mixin that enables field selection based on request arguments."""

    @property
    def fields(self):
        """Filter fields based on request query parameters."""
        fields = super(SelectiveFieldMixin, self).fields
        return apply_subfield_projection(self, copy.copy(fields))



from __future__ import absolute_import, division, print_function, unicode_literals

from django.conf.urls import include, url


urlpatterns = [
    url(r'^', include('resolwe.api_urls')),
    url(r'^api-auth/', include('rest_framework.urls', namespace='rest_framework'))
]



""".. Ignore pydocstyle D400.

=====================
Permissions shortcuts
=====================

.. autofunction:: _group_groups
.. autofunction:: get_object_perms

"""
from __future__ import unicode_literals

from collections import defaultdict
from itertools import chain, groupby

import six

from django.conf import settings
from django.contrib.auth.models import AnonymousUser, Group, Permission
from django.contrib.contenttypes.models import ContentType
from django.db.models import Count, Q
from django.shortcuts import _get_queryset

from guardian.compat import get_user_model
from guardian.ctypes import get_content_type
from guardian.exceptions import MixedContentTypeError, WrongAppError
from guardian.models import GroupObjectPermission
from guardian.shortcuts import get_perms, get_users_with_perms
from guardian.utils import get_anonymous_user, get_group_obj_perms_model, get_identity, get_user_obj_perms_model


# XXX: This is a copy of guardian.shortcuts.get_groups_with_perms with a fixed bug
#      Use the original version once the following fix is merged and released:
#      https://github.com/django-guardian/django-guardian/pull/529
def get_groups_with_perms(obj, attach_perms=False):
    """Return queryset of all ``Group`` objects with *any* object permissions for the given ``obj``."""
    ctype = get_content_type(obj)
    group_model = get_group_obj_perms_model(obj)

    if not attach_perms:
        # It's much easier without attached perms so we do it first if that is the case
        group_rel_name = group_model.group.field.related_query_name()  # pylint: disable=no-member
        if group_model.objects.is_generic():
            group_filters = {
                '%s__content_type' % group_rel_name: ctype,
                '%s__object_pk' % group_rel_name: obj.pk,
            }
        else:
            group_filters = {'%s__content_object' % group_rel_name: obj}
        return Group.objects.filter(**group_filters).distinct()
    else:
        group_perms_mapping = defaultdict(list)
        groups_with_perms = get_groups_with_perms(obj)
        queryset = group_model.objects.filter(group__in=groups_with_perms).prefetch_related('group', 'permission')
        if group_model is GroupObjectPermission:
            queryset = queryset.filter(object_pk=obj.pk, content_type=ctype)
        else:
            queryset = queryset.filter(content_object_id=obj.pk)

        for group_perm in queryset:
            group_perms_mapping[group_perm.group].append(group_perm.permission.codename)
        return dict(group_perms_mapping)


def _group_groups(perm_list):
    """Group permissions by group.

    Input is list of tuples of length 3, where each tuple is in
    following format::

        (<group_id>, <group_name>, <single_permission>)

    Permissions are regrouped and returned in such way that there is
    only one tuple for each group::

        (<group_id>, <group_name>, [<first_permission>, <second_permission>,...])

    :param list perm_list: list of touples of length 3
    :return: list tuples with grouped permissions
    :rtype: list

    """
    perm_list = sorted(perm_list, key=lambda tup: tup[0])

    grouped_perms = []
    for key, group in groupby(perm_list, lambda tup: (tup[0], tup[1])):
        grouped_perms.append((key[0], key[1], [g[2] for g in group]))

    return grouped_perms


def get_user_group_perms(user_or_group, obj):
    """Get permissins for user groups.

    Based on guardian.core.ObjectPermissionChecker.

    """
    user, group = get_identity(user_or_group)

    if user and not user.is_active:
        return [], []
    user_model = get_user_model()
    ctype = ContentType.objects.get_for_model(obj)

    group_model = get_group_obj_perms_model(obj)
    group_rel_name = group_model.permission.field.related_query_name()  # pylint: disable=no-member
    if user:
        user_rel_name = user_model.groups.field.related_query_name()  # pylint: disable=no-member
        group_filters = {user_rel_name: user}
    else:
        group_filters = {'pk': group.pk}
    if group_model.objects.is_generic():
        group_filters.update({
            '{}__content_type'.format(group_rel_name): ctype,
            '{}__object_pk'.format(group_rel_name): obj.pk,
        })
    else:
        group_filters['{}__content_object'.format(group_rel_name)] = obj

    user_perms, group_perms = [], []

    if user:
        perms_qs = Permission.objects.filter(content_type=ctype)
        if user.is_superuser:
            user_perms = list(chain(perms_qs.values_list("codename", flat=True)))
        else:
            model = get_user_obj_perms_model(obj)
            related_name = model.permission.field.related_query_name()  # pylint: disable=no-member
            user_filters = {'{}__user'.format(related_name): user}
            if model.objects.is_generic():
                user_filters.update({
                    '{}__content_type'.format(related_name): ctype,
                    '{}__object_pk'.format(related_name): obj.pk,
                })
            else:
                user_filters['{}__content_object'.format(related_name)] = obj

            user_perms_qs = perms_qs.filter(**user_filters)
            user_perms = list(chain(user_perms_qs.values_list("codename", flat=True)))

    group_perms_qs = Group.objects.filter(**group_filters)
    group_perms = list(chain(group_perms_qs.order_by("pk").values_list(
        "pk", "name", "{}__permission__codename".format(group_rel_name))))

    group_perms = _group_groups(group_perms)

    return user_perms, group_perms


def get_object_perms(obj, user=None):
    """Return permissions for given object in Resolwe specific format.

    Function returns permissions for given object ``obj`` in following
    format::

       {
           "type": "group"/"user"/"public",
           "id": <group_or_user_id>,
           "name": <group_or_user_name>,
           "permissions": [<first_permission>, <second_permission>,...]
       }

    For ``public`` type ``id`` and ``name`` keys are omitted.

    If ``user`` parameter is given, permissions are limited only to
    given user, groups he belongs to and public permissions.

    :param obj: Resolwe's DB model's instance
    :type obj: a subclass of :class:`~resolwe.flow.models.base.BaseModel`
    :param user: Django user
    :type user: :class:`~django.contrib.auth.models.User` or :data:`None`
    :return: list of permissions object in described format
    :rtype: list

    """
    def format_permissions(perms):
        """Remove model name from permission."""
        ctype = ContentType.objects.get_for_model(obj)
        return [perm.replace('_{}'.format(ctype.name), '') for perm in perms]

    perms_list = []

    if user:
        if user.is_authenticated:
            user_perms, group_perms = get_user_group_perms(user, obj)
        else:
            user_perms, group_perms = [], []

        if user_perms != []:
            perms_list.append({
                'type': 'user',
                'id': user.pk,
                'name': user.get_full_name() or user.username,
                'permissions': format_permissions(user_perms),
            })

        if group_perms != []:
            for group_id, group_name, perms in group_perms:
                perms_list.append({
                    'type': 'group',
                    'id': group_id,
                    'name': group_name,
                    'permissions': format_permissions(perms),
                })
    else:
        user_options = {
            'attach_perms': True,
            'with_group_users': False
        }
        for user, perms in six.iteritems(get_users_with_perms(obj, **user_options)):
            if user.username == settings.ANONYMOUS_USER_NAME:
                # public user is treated separately
                continue
            perms_list.append({
                'type': 'user',
                'id': user.pk,
                'name': user.get_full_name() or user.username,
                'permissions': format_permissions(perms),
            })

        group_options = {
            'attach_perms': True,
        }
        for group, perms in six.iteritems(get_groups_with_perms(obj, **group_options)):
            perms_list.append({
                'type': 'group',
                'id': group.pk,
                'name': group.name,
                'permissions': format_permissions(perms),
            })

    public_perms = get_perms(AnonymousUser(), obj)
    if public_perms != []:
        perms_list.append({
            'type': 'public',
            'permissions': format_permissions(public_perms),
        })

    return perms_list


# based on guardian.shortcuts.get_objects_for_user
def get_objects_for_user(user, perms, klass=None, use_groups=True, any_perm=False,
                         with_superuser=True, accept_global_perms=True, perms_filter='pk__in'):
    """Return queryset with required permissions."""
    if isinstance(perms, six.string_types):
        perms = [perms]

    ctype = None
    app_label = None
    codenames = set()

    # Compute codenames set and ctype if possible
    for perm in perms:
        if '.' in perm:
            new_app_label, codename = perm.split('.', 1)
            if app_label is not None and app_label != new_app_label:
                raise MixedContentTypeError(
                    "Given perms must have same app label "
                    "({} != {})".format(app_label, new_app_label))
            else:
                app_label = new_app_label
        else:
            codename = perm
        codenames.add(codename)

        if app_label is not None:
            new_ctype = ContentType.objects.get(app_label=app_label,
                                                permission__codename=codename)
            if ctype is not None and ctype != new_ctype:
                raise MixedContentTypeError(
                    "ContentType was once computed to be {} and another "
                    "one {}".format(ctype, new_ctype))
            else:
                ctype = new_ctype

    # Compute queryset and ctype if still missing
    if ctype is None and klass is not None:
        queryset = _get_queryset(klass)
        ctype = ContentType.objects.get_for_model(queryset.model)
    elif ctype is not None and klass is None:
        queryset = _get_queryset(ctype.model_class())
    elif klass is None:
        raise WrongAppError("Cannot determine content type")
    else:
        queryset = _get_queryset(klass)
        if ctype.model_class() != queryset.model and perms_filter == 'pk__in':
            raise MixedContentTypeError("Content type for given perms and "
                                        "klass differs")

    # At this point, we should have both ctype and queryset and they should
    # match which means: ctype.model_class() == queryset.model
    # we should also have `codenames` list

    # First check if user is superuser and if so, return queryset immediately
    if with_superuser and user.is_superuser:
        return queryset

    # Check if the user is anonymous. The
    # django.contrib.auth.models.AnonymousUser object doesn't work for queries
    # and it's nice to be able to pass in request.user blindly.
    if user.is_anonymous:
        user = get_anonymous_user()

    global_perms = set()
    has_global_perms = False
    # a superuser has by default assigned global perms for any
    if accept_global_perms and with_superuser:
        for code in codenames:
            if user.has_perm(ctype.app_label + '.' + code):
                global_perms.add(code)
        for code in global_perms:
            codenames.remove(code)
        # prerequisite: there must be elements in global_perms otherwise just
        # follow the procedure for object based permissions only AND
        # 1. codenames is empty, which means that permissions are ONLY set
        # globally, therefore return the full queryset.
        # OR
        # 2. any_perm is True, then the global permission beats the object
        # based permission anyway, therefore return full queryset
        if global_perms and (not codenames or any_perm):
            return queryset
        # if we have global perms and still some object based perms differing
        # from global perms and any_perm is set to false, then we have to flag
        # that global perms exist in order to merge object based permissions by
        # user and by group correctly. Scenario: global perm change_xx and
        # object based perm delete_xx on object A for user, and object based
        # permission delete_xx  on object B for group, to which user is
        # assigned.
        # get_objects_for_user(user, [change_xx, delete_xx], use_groups=True,
        # any_perm=False, accept_global_perms=True) must retrieve object A and
        # B.
        elif global_perms and codenames:
            has_global_perms = True

    # Now we should extract list of pk values for which we would filter
    # queryset
    user_model = get_user_obj_perms_model(queryset.model)
    user_obj_perms_queryset = (user_model.objects
                               .filter(Q(user=user) | Q(user=get_anonymous_user()))
                               .filter(permission__content_type=ctype))

    if codenames:
        user_obj_perms_queryset = user_obj_perms_queryset.filter(
            permission__codename__in=codenames)
    direct_fields = ['content_object__pk', 'permission__codename']
    generic_fields = ['object_pk', 'permission__codename']
    if user_model.objects.is_generic():
        user_fields = generic_fields
    else:
        user_fields = direct_fields

    if use_groups:
        group_model = get_group_obj_perms_model(queryset.model)
        group_filters = {
            'permission__content_type': ctype,
            'group__{}'.format(get_user_model().groups.field.related_query_name()): user,  # pylint: disable=no-member
        }
        if codenames:
            group_filters.update({
                'permission__codename__in': codenames,
            })
        groups_obj_perms_queryset = group_model.objects.filter(**group_filters)
        if group_model.objects.is_generic():
            group_fields = generic_fields
        else:
            group_fields = direct_fields
        if not any_perm and codenames and not has_global_perms:
            user_obj_perms = user_obj_perms_queryset.values_list(*user_fields)
            groups_obj_perms = groups_obj_perms_queryset.values_list(*group_fields)
            data = list(user_obj_perms) + list(groups_obj_perms)
            # sorting/grouping by pk (first in result tuple)
            data = sorted(data, key=lambda t: t[0])
            pk_list = []
            for pk, group in groupby(data, lambda t: t[0]):
                obj_codenames = set((e[1] for e in group))
                if codenames.issubset(obj_codenames):
                    pk_list.append(pk)
            objects = queryset.filter(**{perms_filter: pk_list})
            return objects

    if not any_perm and len(codenames) > 1:
        counts = user_obj_perms_queryset.values(
            user_fields[0]).annotate(object_pk_count=Count(user_fields[0]))
        user_obj_perms_queryset = counts.filter(
            object_pk_count__gte=len(codenames))

    values = user_obj_perms_queryset.values_list(user_fields[0], flat=True)
    if user_model.objects.is_generic():
        values = list(values)
    query = Q(**{perms_filter: values})
    if use_groups:
        values = groups_obj_perms_queryset.values_list(group_fields[0], flat=True)
        if group_model.objects.is_generic():
            values = list(values)
        query |= Q(**{perms_filter: values})

    return queryset.filter(query)


def get_users_with_permission(obj, permission):
    """Return users with specific permission on object.

    :param obj: Object to return users for
    :param permission: Permission codename
    """
    user_model = get_user_model()
    return user_model.objects.filter(
        userobjectpermission__object_pk=obj.pk,
        userobjectpermission__permission__codename=permission,
    ).distinct()



"""Overrides for Django settings in tests."""

from __future__ import absolute_import, division, print_function, unicode_literals

import copy

from django.conf import settings

# override all FLOW_EXECUTOR settings that are specified in FLOW_EXECUTOR['TEST']
FLOW_EXECUTOR_SETTINGS = copy.deepcopy(getattr(settings, 'FLOW_EXECUTOR', {}))
TEST_SETTINGS_OVERRIDES = getattr(settings, 'FLOW_EXECUTOR', {}).get('TEST', {})
FLOW_EXECUTOR_SETTINGS.update(TEST_SETTINGS_OVERRIDES)

# override all FLOW_MANAGER settings that are specified in FLOW_MANAGER['TEST']
FLOW_MANAGER_SETTINGS = copy.deepcopy(getattr(settings, 'FLOW_MANAGER', {}))
TEST_SETTINGS_OVERRIDES = getattr(settings, 'FLOW_MANAGER', {}).get('TEST', {})
FLOW_MANAGER_SETTINGS.update(TEST_SETTINGS_OVERRIDES)



""".. Ignore pydocstyle D400.

=================
Resolwe Utilities
=================

"""


class BraceMessage(object):
    """Log messages with the new {}-string formatting syntax.

    .. note::
        When using this helper class, one pays no significant
        performance penalty since the actual formatting only happens
        when (and if) the logged message is actually outputted to a log
        by a handler.

    Example of usage:

        .. code-block:: python

            from resolwe.utils import BraceMessage as __

            logger.error(__("Message with {0} {name}", 2, name="placeholders"))

    Source:
    https://docs.python.org/3/howto/logging-cookbook.html#use-of-alternative-formatting-styles.

    """

    def __init__(self, fmt, *args, **kwargs):
        """Initialize attributes."""
        self.fmt = fmt
        self.args = args
        self.kwargs = kwargs

    def __str__(self):
        """Define the object representation."""
        return self.fmt.format(*self.args, **self.kwargs)



# pylint: disable=missing-docstring
from __future__ import absolute_import, division, print_function, unicode_literals

import os

from asgiref.sync import async_to_sync

from django.db import transaction

from guardian.shortcuts import assign_perm

from resolwe.flow.managers import manager
from resolwe.flow.managers.utils import disable_auto_calls
from resolwe.flow.models import Collection, Data, DataDependency, DescriptorSchema, Process
from resolwe.test import ProcessTestCase, TransactionTestCase

PROCESSES_DIR = os.path.join(os.path.dirname(__file__), 'processes')


class TestManager(ProcessTestCase):

    def setUp(self):
        super(TestManager, self).setUp()

        self.collection = Collection.objects.create(contributor=self.contributor)

        self._register_schemas(path=[PROCESSES_DIR])

    def test_create_data(self):
        """Test that manager is run when new object is created."""
        process = Process.objects.filter(slug='test-min').latest()
        data = Data.objects.create(
            name='Test data',
            contributor=self.contributor,
            process=process,
        )

        data.refresh_from_db()
        self.assertEqual(data.status, Data.STATUS_DONE)

    def test_spawned_process(self):
        """Test that manager is run for spawned processes and permissions are copied."""
        DescriptorSchema.objects.create(
            name='Test schema', slug='test-schema', contributor=self.contributor
        )
        spawned_process = Process.objects.filter(slug='test-save-file').latest()
        # Patch the process to create Entity, so its bahaviour can be tested.
        spawned_process.flow_collection = 'test-schema'
        spawned_process.save()

        process = Process.objects.filter(slug='test-spawn-new').latest()

        with transaction.atomic():
            data = Data.objects.create(
                name='Test data',
                contributor=self.contributor,
                process=process,
            )
            self.collection.data.add(data)
            assign_perm('view_data', self.user, data)

        # Created and spawned objects should be done.
        self.assertEqual(Data.objects.filter(status=Data.STATUS_DONE).count(), 2)

        # Check that permissions are inherited.
        child = Data.objects.last()
        self.assertTrue(self.user.has_perm('view_data', child))
        self.assertEqual(child.collection_set.count(), 1)
        self.assertEqual(child.collection_set.first().pk, self.collection.pk)
        self.assertEqual(child.entity_set.first().collections.count(), 1)
        self.assertEqual(child.entity_set.first().collections.first().pk, self.collection.pk)

    def test_workflow(self):
        """Test that manager is run for workflows."""
        workflow = Process.objects.filter(slug='test-workflow-1').latest()
        data1 = Data.objects.create(name='Test data 1', contributor=self.contributor, process=workflow,
                                    input={'param1': 'world'})
        data2 = Data.objects.create(name='Test data 2', contributor=self.contributor, process=workflow,
                                    input={'param1': 'foobar'})

        # Created and spawned objects should be done.
        self.assertEqual(Data.objects.filter(status=Data.STATUS_DONE).count(), 6)

        # Check correct dependency type is created.
        self.assertEqual({d.kind for d in data1.children_dependency.all()}, {DataDependency.KIND_SUBPROCESS})
        self.assertEqual({d.kind for d in data2.children_dependency.all()}, {DataDependency.KIND_SUBPROCESS})

    def test_dependencies(self):
        """Test that manager handles dependencies correctly."""
        process_parent = Process.objects.filter(slug='test-dependency-parent').latest()
        process_child = Process.objects.filter(slug='test-dependency-child').latest()
        data_parent = Data.objects.create(name='Test parent', contributor=self.contributor,
                                          process=process_parent)
        data_child1 = Data.objects.create(name='Test child', contributor=self.contributor,
                                          process=process_child, input={})
        data_child2 = Data.objects.create(name='Test child', contributor=self.contributor,
                                          process=process_child, input={'parent': data_parent.pk})
        data_child3 = Data.objects.create(name='Test child', contributor=self.contributor,
                                          process=process_child, input={'parent': None})

        data_parent.refresh_from_db()
        data_child1.refresh_from_db()
        data_child2.refresh_from_db()
        data_child3.refresh_from_db()
        self.assertEqual(data_parent.status, Data.STATUS_DONE)
        self.assertEqual(data_child1.status, Data.STATUS_DONE)
        self.assertEqual(data_child2.status, Data.STATUS_DONE)
        self.assertEqual(data_child3.status, Data.STATUS_DONE)

    def test_process_notifications(self):
        process = Process.objects.filter(slug='test-process-notifications').latest()
        data = Data.objects.create(
            name='Test data',
            contributor=self.contributor,
            process=process,
        )

        data.refresh_from_db()

        self.assertEqual(len(data.process_info), 2)
        self.assertEqual(data.process_info[0], 'abc')
        self.assertEqual(data.process_info[1][-5:], 'xx...')

        self.assertEqual(len(data.process_warning), 1)
        self.assertEqual(data.process_warning[0][-5:], 'yy...')

        self.assertEqual(len(data.process_error), 1)
        self.assertEqual(data.process_error[0][-5:], 'zz...')


class TransactionTestManager(TransactionTestCase):

    @disable_auto_calls()
    def test_communicate(self):
        process = Process.objects.create(
            name='Input process',
            contributor=self.contributor,
            type='data:test:',
            input_schema=[
                {
                    'name': 'input_data',
                    'type': 'data:test:',
                    'required': False,
                },
            ]
        )

        data_1 = Data.objects.create(contributor=self.contributor, process=process)
        data_2 = Data.objects.create(contributor=self.contributor, process=process, input={'input_data': data_1.id})
        Data.objects.create(contributor=self.contributor, process=process)
        Data.objects.create(contributor=self.contributor, process=process)

        self.assertEqual(Data.objects.filter(status=Data.STATUS_RESOLVING).count(), 4)

        # Process only one object.
        async_to_sync(manager.communicate)(data_id=data_1.pk, run_sync=True)

        data_1.refresh_from_db()
        self.assertEqual(data_1.status, Data.STATUS_WAITING)
        self.assertEqual(Data.objects.filter(status=Data.STATUS_RESOLVING).count(), 3)

        data_1.status = Data.STATUS_DONE
        data_1.save()

        # Process object's children.
        async_to_sync(manager.communicate)(data_id=data_1.pk, run_sync=True)

        data_2.refresh_from_db()
        self.assertEqual(data_2.status, Data.STATUS_WAITING)
        self.assertEqual(Data.objects.filter(status=Data.STATUS_RESOLVING).count(), 2)

        # Process all objects.
        async_to_sync(manager.communicate)(run_sync=True)

        self.assertEqual(Data.objects.filter(status=Data.STATUS_RESOLVING).count(), 0)



"""Abstract Workflow Engine."""
from __future__ import absolute_import, division, print_function, unicode_literals

import os
import pkgutil
from importlib import import_module

from django.core.exceptions import ImproperlyConfigured
from django.utils._os import upath


class InvalidEngineError(Exception):
    """Raised when an invalid engine is requested."""

    pass


class BaseEngine(object):
    """Base class for all engines in Resolwe workflow."""

    name = None

    def __init__(self, manager, settings=None):
        """Construct a Resolwe engine."""
        self.manager = manager
        self.settings = settings or {}

    def get_name(self):
        """Return the engine name."""
        return self.name


def load_engines(manager, class_name, base_module, engines, class_key='ENGINE', engine_type='engine'):
    """Load engines."""
    loaded_engines = {}

    for module_name_or_dict in engines:
        if not isinstance(module_name_or_dict, dict):
            module_name_or_dict = {
                class_key: module_name_or_dict
            }

        try:
            module_name = module_name_or_dict[class_key]
            engine_settings = module_name_or_dict
        except KeyError:
            raise ImproperlyConfigured("If {} specification is a dictionary, it must define {}".format(
                engine_type, class_key))

        try:
            engine_module = import_module(module_name)
            try:
                engine = getattr(engine_module, class_name)(manager=manager, settings=engine_settings)
                if not isinstance(engine, BaseEngine):
                    raise ImproperlyConfigured("{} module {} class {} must extend BaseEngine".format(
                        engine_type.capitalize(), module_name, class_name))
            except AttributeError:
                raise ImproperlyConfigured("{} module {} is missing a {} class".format(
                    engine_type.capitalize(), module_name, class_name))

            if engine.get_name() in loaded_engines:
                raise ImproperlyConfigured("Duplicated {} {}".format(engine_type, engine.get_name()))

            loaded_engines[engine.get_name()] = engine

        except ImportError as ex:
            # The engine wasn't found. Display a helpful error message listing all possible
            # (built-in) engines.
            engine_dir = os.path.join(os.path.dirname(upath(__file__)), base_module)

            try:
                builtin_engines = [name for _, name, _ in pkgutil.iter_modules([engine_dir])]
            except EnvironmentError:
                builtin_engines = []

            if module_name not in ['resolwe.flow.{}.{}'.format(base_module, builtin_engine)
                                   for builtin_engine in builtin_engines]:
                engine_reprs = map(repr, sorted(builtin_engines))
                error_msg = ("{} isn't an available dataflow {}.\n"
                             "Try using 'resolwe.flow.{}.XXX', where XXX is one of:\n"
                             "    {}\n"
                             "Error was: {}".format(
                                 module_name, engine_type, base_module, ", ".join(engine_reprs), ex
                             ))
                raise ImproperlyConfigured(error_msg)
            else:
                # If there's some other error, this must be an error in Django
                raise

    return loaded_engines



"""Storage viewset."""
from __future__ import absolute_import, division, print_function, unicode_literals

from rest_framework import mixins, viewsets

from resolwe.flow.models import Storage
from resolwe.flow.serializers import StorageSerializer


class StorageViewSet(mixins.RetrieveModelMixin,
                     mixins.ListModelMixin,
                     viewsets.GenericViewSet):
    """API view for :class:`Storage` objects."""

    queryset = Storage.objects.all().prefetch_related('contributor')
    serializer_class = StorageSerializer
    filter_fields = ('contributor', 'name', 'created', 'modified', 'slug')



"""Null workflow executor, preparation portion.

To customize the settings the manager serializes for the executor
at runtime, properly subclass
:class:`~resolwe.flow.executors.prepare.BaseFlowExecutorPreparer` into
``FlowExecutorPreparer`` and override its :meth:`extend_settings`
method.
"""

from ..prepare import BaseFlowExecutorPreparer as FlowExecutorPreparer  # pylint: disable=unused-import



""".. Ignore pydocstyle D400.

============================
Documentation from Processes
============================

Sphinx extension for semi-automatic documentation of Resolwe processes.

This module introduces a new directive that enables user to document a
set of processes. The directive can be used in three variations::

    .. autoprocess::

    .. autoprocess:: category_name

    .. autoprocess:: process_slug

The first option documents *all* processes.

The second option documents just the ones that have *category* equal to
``category_name``. This means that subcategories (e.g. ``analyses:alignment``,
``analyses:variants``) of a category (e.g. ``analyses``) will not be documented.

Third option documents only one process: the one with *slug* equal to
``process_slug``.

"""

import fnmatch
import os
import re
from itertools import groupby
from operator import itemgetter

import six
import yaml
from docutils import nodes
from docutils.parsers.rst import Directive
from sphinx import addnodes

from resolwe.flow.utils import iterate_schema

PROCESS_CACHE = None


def get_process_definition_start(fname, slug):
    """Find the first line of process definition.

    The first line of process definition is the line with a slug.

    :param str fname: Path to filename with processes
    :param string slug: process slug
    :return: line where the process definiton starts
    :rtype: int

    """
    with open(fname) as file_:
        for i, line in enumerate(file_):
            if re.search(r'slug:\s*{}'.format(slug), line):
                return i + 1
    # In case starting line is not found just return first line
    return 1


def get_processes(process_dir, base_source_uri):
    """Find processes in path.

    :param str process_dir: Path to the directory where to search for processes
    :param str base_source_uri: Base URL of the source code repository with process definitions
    :return: Dictionary of processes where keys are URLs pointing to processes'
    source code and values are processes' definitions parsed from YAML files
    :rtype: dict
    :raises: ValueError: if multiple processes with the same slug are found

    """
    global PROCESS_CACHE  # pylint: disable=global-statement
    if PROCESS_CACHE is not None:
        return PROCESS_CACHE

    all_process_files = []
    process_file_extensions = ['*.yaml', '*.yml']
    for root, _, filenames in os.walk(process_dir):
        for extension in process_file_extensions:
            for filename in fnmatch.filter(filenames, extension):
                all_process_files.append(os.path.join(root, filename))

    def read_yaml_file(fname):
        """Read the yaml file."""
        with open(fname) as f:
            return yaml.load(f)

    processes = []
    for process_file in all_process_files:
        processes_in_file = read_yaml_file(process_file)
        for process in processes_in_file:
            # This section finds the line in file where the
            # defintion of the process starts. (there are
            # multiple process definition in some files).
            startline = get_process_definition_start(process_file, process['slug'])

            # Put together URL to starting line of process definition.
            process['source_uri'] = base_source_uri + process_file[len(process_dir) + 1:] + '#L' + str(startline)

            if 'category' not in process:
                process['category'] = 'uncategorized'

            processes.append(process)

    PROCESS_CACHE = processes
    return processes


class AutoProcessDirective(Directive):
    """Automatically document Resolwe processes."""

    has_content = True
    required_arguments = 0
    optional_arguments = 0
    final_argument_whitespace = False
    option_spec = None

    def make_field(self, field_name, field_body):
        """Fill content into nodes.

        :param string field_name: Field name of the field
        :param field_name: Field body if the field
        :type field_name: str or instance of docutils.nodes
        :return: field instance filled with given name and body
        :rtype: nodes.field

        """
        name = nodes.field_name()
        name += nodes.Text(field_name)

        paragraph = nodes.paragraph()
        if isinstance(field_body, six.string_types):
            # This is the case when field_body is just a string:
            paragraph += nodes.Text(field_body)
        else:
            # This is the case when field_body is a complex node:
            # useful when constructing nested field lists
            paragraph += field_body

        body = nodes.field_body()
        body += paragraph

        field = nodes.field()
        field.extend([name, body])
        return field

    def make_properties_list(self, field):
        """Fill the ``field`` into a properties list and return it.

        :param dict field: the content of the property list to make
        :return: field_list instance filled with given field
        :rtype: nodes.field_list

        """
        properties_list = nodes.field_list()

        # changing the order of elements in this list affects
        # the order in which they are displayed
        property_names = ['label', 'type', 'description', 'required',
                          'disabled', 'hidden', 'default', 'placeholder',
                          'validate_regex', 'choices', 'collapse', 'group']

        for name in property_names:
            if name not in field:
                continue

            value = field[name]

            # Value should be formatted in code-style (=literal) mode
            if name in ['type', 'default', 'placeholder', 'validate_regex']:
                literal_node = nodes.literal(str(value), str(value))
                properties_list += self.make_field(name, literal_node)

            # Special formating of ``value`` is needed if name == 'choices'
            elif name == 'choices':
                bullet_list = nodes.bullet_list()
                for choice in value:
                    label = nodes.Text(choice['label'] + ': ')
                    val = nodes.literal(choice['value'], choice['value'])

                    paragraph = nodes.paragraph()
                    paragraph += label
                    paragraph += val
                    list_item = nodes.list_item()
                    list_item += paragraph
                    bullet_list += list_item

                properties_list += self.make_field(name, bullet_list)

            else:
                properties_list += self.make_field(name, str(value))

        return properties_list

    def make_process_header(self, slug, typ, version, source_uri, description, inputs):
        """Generate a process definition header.

        :param str slug: process' slug
        :param str typ: process' type
        :param str version:  process' version
        :param str source_uri: url to the process definition
        :param str description: process' description
        :param dict inputs: process' inputs

        """
        node = addnodes.desc()
        signode = addnodes.desc_signature(slug, '')
        node.append(signode)

        node['objtype'] = node['desctype'] = typ

        signode += addnodes.desc_annotation(typ, typ, classes=['process-type'])
        signode += addnodes.desc_addname('', '')
        signode += addnodes.desc_name(slug + ' ', slug + ' ')

        paramlist = addnodes.desc_parameterlist()

        for field_schema, _, _ in iterate_schema({}, inputs, ''):
            field_type = field_schema['type']
            field_name = field_schema['name']

            field_default = field_schema.get('default', None)
            field_default = '' if field_default is None else '={}'.format(field_default)

            param = addnodes.desc_parameter('', '', noemph=True)
            param += nodes.emphasis(field_type, field_type, classes=['process-type'])
            # separate by non-breaking space in the output
            param += nodes.strong(text=u'\xa0\xa0' + field_name)

            paramlist += param

        signode += paramlist
        signode += nodes.reference('', nodes.Text('[Source: v{}]'.format(version)),
                                   refuri=source_uri, classes=['viewcode-link'])

        desc = nodes.paragraph()
        desc += nodes.Text(description, description)

        return [node, desc]

    def make_process_node(self, process):
        """Fill the content of process definiton node.

        :param dict process: process data as given from yaml.load function
        :return: process node

        """
        name = process['name']
        slug = process['slug']
        typ = process['type']
        version = process['version']
        description = process.get('description', '')
        source_uri = process['source_uri']
        inputs = process.get('input', [])
        outputs = process.get('output', [])

        # Make process name a section title:
        section = nodes.section(ids=['process-' + slug])
        section += nodes.title(name, name)

        # Make process header:
        section += self.make_process_header(slug, typ, version, source_uri, description, inputs)

        # Make inputs section:
        container_node = nodes.container(classes=['toggle'])
        container_header = nodes.paragraph(classes=['header'])
        container_header += nodes.strong(text='Input arguments')
        container_node += container_header

        container_body = nodes.container()
        for field_schema, _, path in iterate_schema({}, inputs, ''):
            container_body += nodes.strong(text=path)
            container_body += self.make_properties_list(field_schema)

        container_node += container_body
        section += container_node

        # Make outputs section:
        container_node = nodes.container(classes=['toggle'])
        container_header = nodes.paragraph(classes=['header'])
        container_header += nodes.strong(text='Output results')
        container_node += container_header

        container_body = nodes.container()
        for field_schema, _, path in iterate_schema({}, outputs, ''):
            container_body += nodes.strong(text=path)
            container_body += self.make_properties_list(field_schema)

        container_node += container_body
        section += container_node

        return [section, addnodes.index(entries=[('single', name, 'process-' + slug, '', None)])]

    def run(self):
        """Create a list of process definitions."""
        config = self.state.document.settings.env.config

        # Get all processes:
        processes = get_processes(config.autoprocess_process_dir, config.autoprocess_source_base_url)
        process_nodes = []

        for process in sorted(processes, key=itemgetter('name')):
            process_nodes.extend(self.make_process_node(process))

        return process_nodes


class AutoProcessCategoryDirective(Directive):
    """Automatically document process categories."""

    has_content = True
    required_arguments = 0
    optional_arguments = 0
    final_argument_whitespace = False
    option_spec = None

    def run(self):
        """Create a category tree."""
        config = self.state.document.settings.env.config

        # Group processes by category
        processes = get_processes(config.autoprocess_process_dir, config.autoprocess_source_base_url)
        processes.sort(key=itemgetter('category'))
        categorized_processes = {k: list(g) for k, g in groupby(processes, itemgetter('category'))}

        # Build category tree
        category_sections = {'': nodes.container(ids=['categories'])}
        top_categories = []

        for category in sorted(categorized_processes.keys()):
            category_path = ''

            for category_node in category.split(':'):
                parent_category_path = category_path
                category_path += '{}:'.format(category_node)

                if category_path in category_sections:
                    continue

                category_name = category_node.capitalize()

                section = nodes.section(ids=['category-' + category_node])
                section += nodes.title(category_name, category_name)

                # Add process list
                category_key = category_path[:-1]
                if category_key in categorized_processes:
                    listnode = nodes.bullet_list()
                    section += listnode

                    for process in categorized_processes[category_key]:
                        par = nodes.paragraph()

                        node = nodes.reference('', process['name'], internal=True)
                        node['refuri'] = config.autoprocess_definitions_uri + '#process-' + process['slug']
                        node['reftitle'] = process['name']

                        par += node
                        listnode += nodes.list_item('', par)

                category_sections[parent_category_path] += section
                category_sections[category_path] = section

                if parent_category_path == '':
                    top_categories.append(section)

        # Return top sections only
        return top_categories


class AutoProcessTypesDirective(Directive):
    """Automatically document process types."""

    has_content = True
    required_arguments = 0
    optional_arguments = 0
    final_argument_whitespace = False
    option_spec = None

    def run(self):
        """Create a type list."""
        config = self.state.document.settings.env.config

        # Group processes by category
        processes = get_processes(config.autoprocess_process_dir, config.autoprocess_source_base_url)
        processes.sort(key=itemgetter('type'))
        processes_by_types = {k: list(g) for k, g in groupby(processes, itemgetter('type'))}

        listnode = nodes.bullet_list()

        for typ in sorted(processes_by_types.keys()):
            par = nodes.paragraph()
            par += nodes.literal(typ, typ)
            par += nodes.Text(' - ')

            processes = sorted(processes_by_types[typ], key=itemgetter('name'))
            last_process = processes[-1]
            for process in processes:
                node = nodes.reference('', process['name'], internal=True)
                node['refuri'] = config.autoprocess_definitions_uri + '#process-' + process['slug']
                node['reftitle'] = process['name']
                par += node
                if process != last_process:
                    par += nodes.Text(', ')

            listnode += nodes.list_item('', par)

        return [listnode]


def setup(app):
    """Register directives.

    When sphinx loads the extension (= imports the extension module) it
    also executes the setup() function. Setup is the way extension
    informs Sphinx about everything that the extension enables: which
    config_values are introduced, which custom nodes/directives/roles
    and which events are defined in extension.

    In this case, only one new directive is created. All used nodes are
    constructed from already existing nodes in docutils.nodes package.

    """
    app.add_config_value('autoprocess_process_dir', '', 'env')
    app.add_config_value('autoprocess_source_base_url', '', 'env')
    app.add_config_value('autoprocess_definitions_uri', '', 'env')

    app.add_directive('autoprocess', AutoProcessDirective)
    app.add_directive('autoprocesscategory', AutoProcessCategoryDirective)
    app.add_directive('autoprocesstype', AutoProcessTypesDirective)

    # The setup() function can return a dictionary. This is treated by
    # Sphinx as metadata of the extension:
    return {'version': '0.2'}



"""Workflow execution engine exceptions."""
from __future__ import absolute_import, division, print_function, unicode_literals


class ExecutionError(Exception):
    """Raised for errors that occur during execution engine evaluation."""

    pass



# pylint: disable=missing-docstring
import os

from resolwe.test import ProcessTestCase, tag_process, with_docker_executor


class FilesProcessTestCase(ProcessTestCase):

    def setUp(self):
        super(FilesProcessTestCase, self).setUp()
        self.files_path = os.path.abspath(os.path.join(os.path.dirname(__file__), 'files'))

    @with_docker_executor
    @tag_process('upload-file')
    def test_upload_file(self):
        upload_file = self.run_process('upload-file', {'src': 'file binary'})
        self.assertFile(upload_file, 'file', 'file binary')

    @with_docker_executor
    @tag_process('upload-image-file')
    def test_upload_image(self):
        upload_image = self.run_process('upload-image-file', {'src': 'file image.png'})
        self.assertFile(upload_image, 'file', 'file image.png')

    @with_docker_executor
    @tag_process('upload-tab-file')
    def test_upload_tabular_tab(self):
        upload_tab = self.run_process('upload-tab-file', {'src': 'file tab.tab.gz'})
        self.assertFile(upload_tab, 'file', 'tab_file_tabular.tab.gz', compression='gzip')
        self.assertFile(upload_tab, 'src_file', 'file tab.tab.gz', compression='gzip')

    @with_docker_executor
    @tag_process('upload-tab-file')
    def test_upload_tabular_csv(self):
        upload_csv = self.run_process('upload-tab-file', {'src': 'file tab.csv.gz'})
        self.assertFile(upload_csv, 'file', 'csv_file_tabular.tab.gz', compression='gzip')
        self.assertFile(upload_csv, 'src_file', 'file tab.csv.gz', compression='gzip')

    @with_docker_executor
    @tag_process('upload-tab-file')
    def test_upload_tabular_xls(self):
        upload_xls = self.run_process('upload-tab-file', {'src': 'file tab.xls'})
        self.assertFile(upload_xls, 'file', 'xls_file_tabular.tab.gz', compression='gzip')
        self.assertFile(upload_xls, 'src_file', 'file tab.xls')

    @with_docker_executor
    @tag_process('upload-tab-file')
    def test_upload_tabular_xlsx(self):
        upload_xlsx = self.run_process('upload-tab-file', {'src': 'file tab.1.xlsx'})
        self.assertFile(upload_xlsx, 'file', 'xlsx_file_tabular.tab.gz', compression='gzip')
        self.assertFile(upload_xlsx, 'src_file', 'file tab.1.xlsx')



# -*- coding: utf-8 -*-
# Generated by Django 1.9.11 on 2017-01-24 05:07
from __future__ import unicode_literals

from django.db import migrations
import versionfield


class Migration(migrations.Migration):

    dependencies = [
        ('flow', '0015_make_data_indexes'),
    ]

    operations = [
        migrations.AlterField(
            model_name='collection',
            name='version',
            field=versionfield.VersionField(default='0.0.0'),
        ),
        migrations.AlterField(
            model_name='data',
            name='version',
            field=versionfield.VersionField(default='0.0.0'),
        ),
        migrations.AlterField(
            model_name='descriptorschema',
            name='version',
            field=versionfield.VersionField(default='0.0.0'),
        ),
        migrations.AlterField(
            model_name='entity',
            name='version',
            field=versionfield.VersionField(default='0.0.0'),
        ),
        migrations.AlterField(
            model_name='process',
            name='version',
            field=versionfield.VersionField(default='0.0.0'),
        ),
        migrations.AlterField(
            model_name='storage',
            name='version',
            field=versionfield.VersionField(default='0.0.0'),
        ),
        migrations.AlterField(
            model_name='trigger',
            name='version',
            field=versionfield.VersionField(default='0.0.0'),
        ),
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.9.7 on 2016-10-27 05:37
from __future__ import unicode_literals

import django.contrib.postgres.fields.jsonb
from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('flow', '0012_require_checksum'),
    ]

    operations = [
        migrations.AddField(
            model_name='process',
            name='requirements',
            field=django.contrib.postgres.fields.jsonb.JSONField(default=dict),
        ),
    ]



# pylint: disable=missing-docstring
from __future__ import absolute_import, division, print_function, unicode_literals

from guardian.models import GroupObjectPermission, UserObjectPermission
from guardian.shortcuts import assign_perm

from resolwe.flow.models import Collection, DescriptorSchema, Process
from resolwe.permissions.utils import change_perm_ctype, copy_permissions, get_full_perm, get_perm_action
from resolwe.test import TestCase


class UtilsTestCase(TestCase):
    def setUp(self):
        super(UtilsTestCase, self).setUp()

        self.src_process = Process.objects.create(name='Source process', contributor=self.contributor)
        self.dst_process = Process.objects.create(name='Destination process', contributor=self.contributor)

        self.collection = Collection.objects.create(name='Test collection', contributor=self.contributor)

        self.descriptor_schema = DescriptorSchema.objects.create(contributor=self.contributor)

    def test_get_perm_action(self):
        self.assertEqual(get_perm_action('view_data'), 'view')
        self.assertEqual(get_perm_action('view'), 'view')

    def test_get_full_perm(self):
        self.assertEqual(get_full_perm('view', self.collection), 'view_collection')
        self.assertEqual(get_full_perm('view', self.descriptor_schema), 'view_descriptorschema')

    def test_change_perm_ctype(self):
        self.assertEqual(change_perm_ctype('view_data', self.collection), 'view_collection')

    def test_copy_permissions(self):
        assign_perm('view_process', self.contributor, self.src_process)
        assign_perm('view_process', self.group, self.src_process)

        copy_permissions(self.src_process, self.dst_process)

        self.assertEqual(GroupObjectPermission.objects.count(), 2)
        self.assertEqual(UserObjectPermission.objects.count(), 2)

        self.assertTrue(self.contributor.has_perm('flow.view_process', self.dst_process))
        # User inherites permission from group
        self.assertTrue(self.user.has_perm('flow.view_process', self.dst_process))

    def test_copy_different_ctype(self):
        assign_perm('view_collection', self.contributor, self.collection)
        assign_perm('add_collection', self.contributor, self.collection)

        copy_permissions(self.collection, self.dst_process)

        # Only 'view' is copied as process has no 'add' permission.
        self.assertEqual(UserObjectPermission.objects.count(), 3)



# pylint: disable=missing-docstring
from __future__ import absolute_import, division, print_function, unicode_literals

import io
import json
import os
import shutil

import six
from mock import MagicMock, patch

from django.conf import settings
from django.contrib.auth import get_user_model
from django.core.exceptions import ValidationError
from django.db import transaction

from guardian.shortcuts import assign_perm, remove_perm
from rest_framework.test import APIRequestFactory, APITestCase, force_authenticate

from resolwe.flow.expression_engines import EvaluationError
from resolwe.flow.models import Data, DataDependency, DescriptorSchema, Entity, Process, Storage
from resolwe.flow.models.data import hydrate_size, render_template
from resolwe.flow.models.utils import hydrate_input_references
from resolwe.flow.views import DataViewSet
from resolwe.test import TestCase, TransactionTestCase

try:
    import builtins  # py3
except ImportError:
    import __builtin__ as builtins  # py2


class DataModelNameTest(TransactionTestCase):

    def test_name(self):
        process = Process.objects.create(slug='test-first',
                                         type='data:test:first:',
                                         contributor=self.contributor,
                                         data_name='Process based data name',
                                         output_schema=[{
                                             'name': 'stat',
                                             'type': 'basic:string:',
                                             'required': False,
                                         }],
                                         run={
                                             'language': 'bash',
                                             'program': 'echo {"stat": "42"}'
                                         })

        data = Data.objects.create(contributor=self.contributor,
                                   process=process)
        data.refresh_from_db()

        self.assertEqual(data.name, 'Process based data name')
        self.assertFalse(data.named_by_user)

        data.name = 'Name changed by user'
        data.save()
        data.refresh_from_db()

        self.assertEqual(data.name, 'Name changed by user')
        self.assertTrue(data.named_by_user)

        data = Data.objects.create(name='Explicit data name',
                                   contributor=self.contributor,
                                   process=process)
        data.refresh_from_db()

        self.assertEqual(data.name, 'Explicit data name')
        self.assertTrue(data.named_by_user)

        process = Process.objects.create(slug='test-second',
                                         type='test:second',
                                         contributor=self.contributor,
                                         requirements={'expression-engine': 'jinja'},
                                         data_name='Process based data name, value: {{src.stat}}',
                                         input_schema=[{
                                             'name': 'src',
                                             'type': 'data:test:first:',
                                             'required': False,
                                         }])

        with transaction.atomic():
            second = Data.objects.create(contributor=self.contributor,
                                         process=process,
                                         input={'src': data.id})

            data.output = {'stat': '42'}
            data.status = 'OK'
            data.save()

            self.assertEqual(second.name, 'Process based data name, value: ')
            self.assertFalse(second.named_by_user)

        second = Data.objects.get(id=second.id)

        self.assertEqual(second.name, 'Process based data name, value: 42')
        self.assertFalse(second.named_by_user)

        with transaction.atomic():
            data.output = {}
            data.status = 'RE'
            data.save()

            second = Data.objects.create(contributor=self.contributor,
                                         process=process,
                                         input={'src': data.id})

            second.name = 'User\' data name'
            second.save()

            data.output = {'stat': '42'}
            data.status = 'OK'
            data.save()

            self.assertEqual(second.name, 'User\' data name')
            self.assertTrue(second.named_by_user)

        second = Data.objects.get(id=second.id)

        self.assertEqual(second.name, 'User\' data name')
        self.assertTrue(second.named_by_user)


class DataModelTest(TestCase):

    def test_delete_chunked(self):
        process = Process.objects.create(contributor=self.contributor)
        for _ in range(1000):
            Data.objects.create(contributor=self.contributor, process=process)

        Data.objects.all().delete_chunked(chunk_size=50)
        self.assertFalse(Data.objects.exists())

    def test_trim_name(self):
        process = Process.objects.create(contributor=self.contributor, data_name='test' * 50)
        data = Data.objects.create(contributor=self.contributor, process=process)

        self.assertEqual(len(data.name), 100)
        self.assertEqual(data.name[-3:], '...')

    def test_hydrate_file_size(self):
        proc = Process.objects.create(
            name='Test process',
            contributor=self.contributor,
            output_schema=[
                {'name': 'output_file', 'type': 'basic:file:'}
            ]
        )

        data = Data.objects.create(
            name='Test data',
            contributor=self.contributor,
            process=proc,
        )

        data.output = {
            'output_file': {'file': 'output.txt'}
        }

        with self.assertRaises(ValidationError):
            data.save()

        dir_path = os.path.join(settings.FLOW_EXECUTOR['DATA_DIR'], str(data.pk))
        os.makedirs(dir_path)
        file_path = os.path.join(dir_path, 'output.txt')
        with open(file_path, 'w') as fn:
            fn.write('foo bar')

        data.save()

        self.assertEqual(data.output['output_file']['size'], 7)

    def test_dependencies_single(self):
        process = Process.objects.create(slug='test-dependencies',
                                         type='data:test:dependencies:',
                                         contributor=self.contributor,
                                         input_schema=[{
                                             'name': 'src',
                                             'type': 'data:test:dependencies:',
                                             'required': False,
                                         }])

        first = Data.objects.create(name='First data',
                                    contributor=self.contributor,
                                    process=process)

        second = Data.objects.create(name='Second data',
                                     contributor=self.contributor,
                                     process=process,
                                     input={'src': first.id})

        third = Data.objects.create(name='Third data',
                                    contributor=self.contributor,
                                    process=process,
                                    input={'src': first.id})

        self.assertEqual(first.parents.all().count(), 0)
        self.assertEqual(first.children.all().count(), 2)
        self.assertEqual(second.children.all().count(), 0)
        self.assertEqual(second.parents.all().count(), 1)
        self.assertEqual(third.children.all().count(), 0)
        self.assertEqual(third.parents.all().count(), 1)
        self.assertIn(first, second.parents.all())
        self.assertIn(first, third.parents.all())
        self.assertIn(second, first.children.all())
        self.assertIn(third, first.children.all())

        # Check correct dependency type is created.
        self.assertEqual({d.kind for d in first.children_dependency.all()}, {DataDependency.KIND_IO})
        self.assertEqual({d.kind for d in second.parents_dependency.all()}, {DataDependency.KIND_IO})
        self.assertEqual({d.kind for d in third.parents_dependency.all()}, {DataDependency.KIND_IO})

    def test_dependencies_list(self):
        process = Process.objects.create(slug='test-dependencies-list',
                                         type='data:test:dependencies:list:',
                                         contributor=self.contributor,
                                         input_schema=[{
                                             'name': 'src',
                                             'type': 'list:data:test:dependencies:list:',
                                             'required': False,
                                         }])

        first = Data.objects.create(name='First data',
                                    contributor=self.contributor,
                                    process=process)

        second = Data.objects.create(name='Second data',
                                     contributor=self.contributor,
                                     process=process,
                                     input={'src': [first.id]})

        third = Data.objects.create(name='Third data',
                                    contributor=self.contributor,
                                    process=process,
                                    input={'src': [first.id, second.id]})

        self.assertEqual(first.parents.all().count(), 0)
        self.assertEqual(first.children.all().count(), 2)
        self.assertEqual(second.children.all().count(), 1)
        self.assertEqual(second.parents.all().count(), 1)
        self.assertEqual(third.children.all().count(), 0)
        self.assertEqual(third.parents.all().count(), 2)
        self.assertIn(first, second.parents.all())
        self.assertIn(first, third.parents.all())
        self.assertIn(second, first.children.all())
        self.assertIn(third, first.children.all())
        self.assertIn(third, second.children.all())

        # Check correct dependency type is created.
        self.assertEqual({d.kind for d in first.children_dependency.all()}, {DataDependency.KIND_IO})
        self.assertEqual({d.kind for d in second.children_dependency.all()}, {DataDependency.KIND_IO})
        self.assertEqual({d.kind for d in second.parents_dependency.all()}, {DataDependency.KIND_IO})
        self.assertEqual({d.kind for d in third.parents_dependency.all()}, {DataDependency.KIND_IO})


class EntityModelTest(TestCase):

    def setUp(self):
        super(EntityModelTest, self).setUp()

        DescriptorSchema.objects.create(name='Sample', slug='sample', contributor=self.contributor)
        self.process = Process.objects.create(name='Test process',
                                              contributor=self.contributor,
                                              flow_collection='sample')
        # `Sample`is created automatically when `Data` object is created
        self.data = Data.objects.create(name='Test data', contributor=self.contributor, process=self.process)

    def test_delete_data(self):
        # Create another Data object and add it to the same Entity.
        data_2 = Data.objects.create(name='Test data', contributor=self.contributor, process=self.process)
        data_2.entity_set.all().delete()
        self.data.entity_set.first().data.add(data_2)

        data_2.delete()
        self.assertEqual(Entity.objects.count(), 1)

        self.data.delete()
        self.assertEqual(Entity.objects.count(), 0)

    def test_two_entities(self):
        entity_2 = Entity.objects.create(contributor=self.contributor)
        entity_2.data.add(self.data)

        self.data.delete()
        self.assertEqual(Entity.objects.count(), 0)

    def test_new_sample(self):
        data = Data.objects.create(name='Test data', contributor=self.contributor, process=self.process,
                                   tags=['foo', 'bar'])
        entity = Entity.objects.last()
        self.assertTrue(entity.data.filter(pk=data.pk).exists())

        # Make sure tags are copied.
        self.assertEqual(entity.tags, data.tags)


class GetOrCreateTestCase(APITestCase):

    def setUp(self):
        super(GetOrCreateTestCase, self).setUp()

        user_model = get_user_model()
        self.user = user_model.objects.create(username='test_user', password='test_pwd')

        self.process = Process.objects.create(
            name='Temporary process',
            contributor=self.user,
            slug='tmp-process',
            persistence=Process.PERSISTENCE_TEMP,
            input_schema=[
                {'name': 'some_value', 'type': 'basic:integer:', 'default': 42}
            ],
        )
        assign_perm('view_process', self.user, self.process)

        process_2 = Process.objects.create(
            name='Another process',
            contributor=self.user,
            slug='another-process',
            persistence=Process.PERSISTENCE_TEMP,
            input_schema=[
                {'name': 'some_value', 'type': 'basic:integer:'}
            ],
        )
        assign_perm('view_process', self.user, process_2)

        self.data = Data.objects.create(
            name='Temporary data',
            contributor=self.user,
            process=self.process,
            input={'some_value': 42}
        )
        assign_perm('view_data', self.user, self.data)

        self.get_or_create_view = DataViewSet.as_view({'post': 'get_or_create'})
        self.factory = APIRequestFactory()

    def tearDown(self):
        for data in Data.objects.all():
            data_dir = os.path.join(settings.FLOW_EXECUTOR['DATA_DIR'], str(data.id))
            shutil.rmtree(data_dir, ignore_errors=True)

        super(GetOrCreateTestCase, self).tearDown()

    def test_get_same(self):
        request = self.factory.post(
            '',
            {'name': 'Data object', 'input': {'some_value': 42}, 'process': 'tmp-process'},
            format='json'
        )
        force_authenticate(request, user=self.user)

        response = self.get_or_create_view(request)
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.data['id'], self.data.pk)

    def test_use_defaults(self):
        request = self.factory.post(
            '',
            {'name': 'Data object', 'input': {}, 'process': 'tmp-process'},
            format='json'
        )
        force_authenticate(request, user=self.user)

        response = self.get_or_create_view(request)
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.data['id'], self.data.pk)

    def test_missing_permission(self):
        remove_perm('view_data', self.user, self.data)

        request = self.factory.post(
            '',
            {'name': 'Data object', 'input': {'some_value': 42}, 'process': 'tmp-process'},
            format='json'
        )
        force_authenticate(request, user=self.user)

        response = self.get_or_create_view(request)
        self.assertEqual(response.status_code, 201)
        self.assertNotEqual(response.data['id'], self.data.pk)

    def test_different_input(self):
        request = self.factory.post(
            '',
            {'name': 'Data object', 'input': {'some_value': 43}, 'process': 'tmp-process'},
            format='json'
        )
        force_authenticate(request, user=self.user)

        response = self.get_or_create_view(request)
        self.assertEqual(response.status_code, 201)
        self.assertNotEqual(response.data['id'], self.data.pk)

    def test_different_process(self):
        request = self.factory.post(
            '',
            {'name': 'Data object', 'input': {'some_value': 43}, 'process': 'another-process'},
            format='json'
        )
        force_authenticate(request, user=self.user)

        response = self.get_or_create_view(request)
        self.assertEqual(response.status_code, 201)
        self.assertNotEqual(response.data['id'], self.data.pk)

    def test_different_process_version(self):
        self.process.version = '2.0.0'
        self.process.save()

        request = self.factory.post(
            '',
            {'name': 'Data object', 'input': {'some_value': 42}, 'process': 'tmp-process'},
            format='json'
        )
        force_authenticate(request, user=self.user)

        response = self.get_or_create_view(request)
        self.assertEqual(response.status_code, 201)
        self.assertNotEqual(response.data['id'], self.data.pk)

    def test_raw_process(self):
        self.process.persistence = Process.PERSISTENCE_RAW
        self.process.save()

        request = self.factory.post(
            '',
            {'name': 'Data object', 'input': {'some_value': 42}, 'process': 'tmp-process'},
            format='json'
        )
        force_authenticate(request, user=self.user)

        response = self.get_or_create_view(request)
        self.assertEqual(response.status_code, 201)
        self.assertNotEqual(response.data['id'], self.data.pk)


@patch('resolwe.flow.models.utils.os')
class HydrateFileSizeUnitTest(TestCase):

    def setUp(self):
        super(HydrateFileSizeUnitTest, self).setUp()

        self.process = Process(
            output_schema=[
                {'name': 'test_file', 'type': 'basic:file:', 'required': False},
                {'name': 'file_list', 'type': 'list:basic:file:', 'required': False}
            ]
        )
        self.data = Data(
            pk=13,
            process=self.process,
            output={'test_file': {'file': 'test_file.tmp'}}
        )

    def test_done_data(self, os_mock):
        os_mock.path.isfile.return_value = True
        os_mock.path.getsize.return_value = 42000
        hydrate_size(self.data)
        self.assertEqual(self.data.output['test_file']['size'], 42000)

    def test_data_with_refs(self, os_mock):
        os_mock.path.isfile.return_value = True
        os_mock.path.getsize.side_effect = [42000, 8000, 42]
        self.data.output = {
            'test_file': {
                'file': 'test_file.tmp',
                'refs': ['ref_file1.tmp', 'ref_file2.tmp']
            }
        }
        hydrate_size(self.data)
        self.assertEqual(self.data.output['test_file']['size'], 42000)
        self.assertEqual(self.data.output['test_file']['total_size'], 50042)
        self.assertEqual(self.data.size, 50042)

    def test_list(self, os_mock):
        os_mock.path.isfile.return_value = True
        os_mock.path.getsize.side_effect = [34, 42000, 42]
        self.data.output = {
            'file_list': [
                {
                    'file': 'test_01.tmp',
                },
                {
                    'file': 'test_02.tmp',
                    'refs': ['ref_file1.tmp']
                },
            ]
        }
        hydrate_size(self.data)
        self.assertEqual(self.data.output['file_list'][0]['size'], 34)
        self.assertEqual(self.data.output['file_list'][0]['total_size'], 34)
        self.assertEqual(self.data.output['file_list'][1]['size'], 42000)
        self.assertEqual(self.data.output['file_list'][1]['total_size'], 42042)
        self.assertEqual(self.data.size, 34 + 42042)

    def test_change_size(self, os_mock):
        """Size is not changed after object is done."""
        os_mock.path.isfile.return_value = True
        os_mock.path.getsize.return_value = 42000
        hydrate_size(self.data)
        self.assertEqual(self.data.output['test_file']['size'], 42000)

        os_mock.path.getsize.return_value = 43000
        hydrate_size(self.data)
        self.assertEqual(self.data.output['test_file']['size'], 43000)

        self.data.status = Data.STATUS_DONE
        os_mock.path.getsize.return_value = 44000
        hydrate_size(self.data)
        self.assertEqual(self.data.output['test_file']['size'], 43000)

    def test_missing_file(self, os_mock):
        os_mock.path.isfile.return_value = False
        with self.assertRaises(ValidationError):
            hydrate_size(self.data)


class StorageModelTestCase(TestCase):

    def setUp(self):
        super(StorageModelTestCase, self).setUp()

        self.proc = Process.objects.create(
            name='Test process',
            contributor=self.contributor,
            output_schema=[
                {'name': 'json_field', 'type': 'basic:json:'},
            ],
        )

    def test_save_storage(self):
        """`basic:json:` fields are stored in Storage"""
        data = Data.objects.create(
            name='Test data',
            contributor=self.contributor,
            process=self.proc,
        )

        data.output = {'json_field': {'foo': 'bar'}}
        data.status = Data.STATUS_DONE
        data.save()

        self.assertEqual(Storage.objects.count(), 1)
        storage = Storage.objects.first()
        self.assertEqual(data.output['json_field'], storage.pk)

    def test_save_storage_file(self):
        """File is loaded and saved to storage"""
        data = Data.objects.create(
            name='Test data',
            contributor=self.contributor,
            process=self.proc,
        )

        data.output = {'json_field': 'json.txt'}
        data.status = Data.STATUS_DONE

        # NOTE: io.StringIO expects a Python 3 (unicode) string
        json_file = io.StringIO(six.u(json.dumps({'foo': 'bar'})))

        isfile_mock = MagicMock(return_value=True)
        open_mock = MagicMock(return_value=json_file)
        with patch.object(os.path, 'isfile', isfile_mock):
            with patch.object(builtins, 'open', open_mock):
                data.save()

        self.assertEqual(Storage.objects.count(), 1)
        storage = Storage.objects.first()
        self.assertEqual(data.output['json_field'], storage.pk)
        self.assertEqual(storage.json, {'foo': 'bar'})

    def test_delete_data(self):
        """`Storage` is deleted when `Data` object is deleted"""
        data = Data.objects.create(
            name='Test data',
            contributor=self.contributor,
            process=self.proc,
        )

        data.output = {'json_field': {'foo': 'bar'}}
        data.status = Data.STATUS_DONE
        data.save()

        self.assertEqual(Storage.objects.count(), 1)

        data.delete()
        self.assertEqual(Storage.objects.count(), 0)

    def test_storage_manager(self):
        data = Data.objects.create(
            name='Test data',
            contributor=self.contributor,
            process=self.proc,
        )

        data.output = {'json_field': {'foo': {'moo': 'bar'}}}
        data.status = Data.STATUS_DONE
        data.save()

        # Annotation with specific JSON field.
        storage = Storage.objects.with_json_path(['foo', 'moo'])[0]
        self.assertEqual(storage.json_foo_moo, 'bar')

        storage = Storage.objects.with_json_path('foo.moo')[0]
        self.assertEqual(storage.json_foo_moo, 'bar')

        storage = Storage.objects.with_json_path(['foo', 'moo'], field='result')[0]
        self.assertEqual(storage.result, 'bar')

        # Extract specific JSON path.
        value = Storage.objects.get_json_path(['foo', 'moo'])[0]
        self.assertEqual(value, 'bar')

        value = Storage.objects.get_json_path('foo.moo')[0]
        self.assertEqual(value, 'bar')


class UtilsTestCase(TestCase):

    def test_render_template(self):
        process_mock = MagicMock(requirements={'expression-engine': 'jinja'})
        template = render_template(process_mock, '{{ 1 | increase }}', {})
        self.assertEqual(template, '2')

    def test_render_template_error(self):
        process_mock = MagicMock(requirements={'expression-engine': 'jinja'})
        with self.assertRaises(EvaluationError):
            render_template(process_mock, '{{ 1 | missing_increase }}', {})

    def test_hydrate_input_references(self):
        process = Process.objects.create(
            contributor=self.contributor,
            type='data:test:',
            output_schema=[
                {
                    'name': 'file',
                    'type': 'basic:file:',
                }, {
                    'name': 'file_list',
                    'type': 'list:basic:file:',
                }, {
                    'name': 'dir',
                    'type': 'basic:dir:',
                }, {
                    'name': 'dir_list',
                    'type': 'list:basic:dir:',
                },
            ],
        )
        descriptor_schema = DescriptorSchema.objects.create(
            contributor=self.contributor,
            schema=[
                {
                    'name': 'annotation',
                    'type': 'basic:string:',
                },
            ],
        )
        data = Data.objects.create(
            contributor=self.contributor,
            status=Data.STATUS_ERROR,
            process=process,
            # Workaround for skipping the validation.
            output={
                'file': {'file': 'some-file', 'refs': ['ref1']},
                'file_list': [{'file': 'some-file', 'refs': ['ref2']}, {'file': 'another-file'}],
                'dir': {'dir': 'some-dir', 'refs': ['ref3']},
                'dir_list': [{'dir': 'some-dir', 'refs': ['ref4']}, {'dir': 'another-dir'}],
            },
            descriptor_schema=descriptor_schema,
            descriptor={
                'annotation': 'my-annotation',
            },
            size=0,
        )

        input_schema = [
            {
                'name': 'data',
                'type': 'data:test:',
            },
        ]
        input_ = {'data': data.pk}
        hydrate_input_references(input_, input_schema)

        path_prefix = os.path.join(settings.FLOW_EXECUTOR['DATA_DIR'], str(data.id))

        self.assertEqual(input_['data']['__descriptor'], {'annotation': 'my-annotation'})
        self.assertEqual(input_['data']['__type'], 'data:test:')
        self.assertEqual(input_['data']['__id'], data.id)

        self.assertEqual(input_['data']['file']['file'].data_id, data.id)
        self.assertEqual(input_['data']['file']['file'].file_name, 'some-file')
        self.assertEqual(str(input_['data']['file']['file']), os.path.join(path_prefix, 'some-file'))

        self.assertEqual(input_['data']['file']['refs'][0].data_id, data.id)
        self.assertEqual(input_['data']['file']['refs'][0].file_name, 'ref1')
        self.assertEqual(str(input_['data']['file']['refs'][0]), os.path.join(path_prefix, 'ref1'))

        self.assertEqual(input_['data']['file_list'][0]['file'].data_id, data.id)
        self.assertEqual(input_['data']['file_list'][0]['file'].file_name, 'some-file')
        self.assertEqual(str(input_['data']['file_list'][0]['file']), os.path.join(path_prefix, 'some-file'))

        self.assertEqual(input_['data']['file_list'][0]['refs'][0].data_id, data.id)
        self.assertEqual(input_['data']['file_list'][0]['refs'][0].file_name, 'ref2')
        self.assertEqual(str(input_['data']['file_list'][0]['refs'][0]), os.path.join(path_prefix, 'ref2'))

        self.assertEqual(input_['data']['file_list'][1]['file'].data_id, data.id)
        self.assertEqual(input_['data']['file_list'][1]['file'].file_name, 'another-file')
        self.assertEqual(str(input_['data']['file_list'][1]['file']), os.path.join(path_prefix, 'another-file'))

        self.assertEqual(input_['data']['dir']['dir'].data_id, data.id)
        self.assertEqual(input_['data']['dir']['dir'].file_name, 'some-dir')
        self.assertEqual(str(input_['data']['dir']['dir']), os.path.join(path_prefix, 'some-dir'))

        self.assertEqual(input_['data']['dir']['refs'][0].data_id, data.id)
        self.assertEqual(input_['data']['dir']['refs'][0].file_name, 'ref3')
        self.assertEqual(str(input_['data']['dir']['refs'][0]), os.path.join(path_prefix, 'ref3'))

        self.assertEqual(input_['data']['dir_list'][0]['dir'].data_id, data.id)
        self.assertEqual(input_['data']['dir_list'][0]['dir'].file_name, 'some-dir')
        self.assertEqual(str(input_['data']['dir_list'][0]['dir']), os.path.join(path_prefix, 'some-dir'))

        self.assertEqual(input_['data']['dir_list'][0]['refs'][0].data_id, data.id)
        self.assertEqual(input_['data']['dir_list'][0]['refs'][0].file_name, 'ref4')
        self.assertEqual(str(input_['data']['dir_list'][0]['refs'][0]), os.path.join(path_prefix, 'ref4'))

        self.assertEqual(input_['data']['dir_list'][1]['dir'].data_id, data.id)
        self.assertEqual(input_['data']['dir_list'][1]['dir'].file_name, 'another-dir')
        self.assertEqual(str(input_['data']['dir_list'][1]['dir']), os.path.join(path_prefix, 'another-dir'))



"""Workflow execution engines."""
from __future__ import absolute_import, division, print_function, unicode_literals

from resolwe.flow.engine import BaseEngine


class BaseExecutionEngine(BaseEngine):
    """A workflow execution engine."""

    def evaluate(self, data):
        """Return the code needed to compute a given Data object."""
        raise NotImplementedError

    def get_expression_engine(self, name):
        """Return an expression engine by its name."""
        return self.manager.get_expression_engine(name)

    def get_output_schema(self, process):
        """Return any additional output schema for the process."""
        return []



"""Workload management system connectors.

The connectors are used as glue between the Resolwe Manager and various
concrete workload management systems that might be used by it. Since the
only functional requirement is job submission, they can be simple and
nearly contextless.
"""



"""Collection viewset."""
from __future__ import absolute_import, division, print_function, unicode_literals

from distutils.util import strtobool  # pylint: disable=import-error,no-name-in-module

from django.db.models.query import Prefetch

from rest_framework import exceptions, mixins, status, viewsets
from rest_framework.decorators import detail_route
from rest_framework.response import Response

from resolwe.flow.filters import CollectionFilter
from resolwe.flow.models import Collection, Data
from resolwe.flow.serializers import CollectionSerializer
from resolwe.permissions.loader import get_permissions_class
from resolwe.permissions.mixins import ResolwePermissionsMixin
from resolwe.permissions.utils import remove_permission, update_permission

from .mixins import ResolweCheckSlugMixin, ResolweCreateModelMixin, ResolweUpdateModelMixin


class CollectionViewSet(ResolweCreateModelMixin,
                        mixins.RetrieveModelMixin,
                        ResolweUpdateModelMixin,
                        mixins.DestroyModelMixin,
                        mixins.ListModelMixin,
                        ResolwePermissionsMixin,
                        ResolweCheckSlugMixin,
                        viewsets.GenericViewSet):
    """API view for :class:`Collection` objects."""

    queryset = Collection.objects.all().prefetch_related(
        'descriptor_schema',
        'contributor',
        Prefetch('data', queryset=Data.objects.all().order_by('id'))
    )
    serializer_class = CollectionSerializer
    permission_classes = (get_permissions_class(),)
    filter_class = CollectionFilter
    ordering_fields = ('id', 'created', 'modified', 'name')
    ordering = ('id',)

    def set_content_permissions(self, user, obj, payload):
        """Apply permissions to data objects and entities in ``Collection``."""
        for entity in obj.entity_set.all():
            if user.has_perm('share_entity', entity):
                update_permission(entity, payload)

        # Data doesn't have "ADD" permission, so it has to be removed
        payload = remove_permission(payload, 'add')

        for data in obj.data.all():
            if user.has_perm('share_data', data):
                update_permission(data, payload)

    def create(self, request, *args, **kwargs):
        """Only authenticated usesr can create new collections."""
        if not request.user.is_authenticated:
            raise exceptions.NotFound

        return super(CollectionViewSet, self).create(request, *args, **kwargs)

    def destroy(self, request, *args, **kwargs):
        """Destroy a model instance.

        If ``delete_content`` flag is set in query parameters, also all
        Data objects and Entities, on which user has ``EDIT``
        permission, contained in collection will be deleted.
        """
        obj = self.get_object()
        user = request.user

        if strtobool(request.query_params.get('delete_content', 'false')):
            for entity in obj.entity_set.all():
                if user.has_perm('edit_entity', entity):
                    entity.delete()

            for data in obj.data.all():
                if user.has_perm('edit_data', data):
                    data.delete()

        return super(CollectionViewSet, self).destroy(request, *args, **kwargs)  # pylint: disable=no-member

    @detail_route(methods=[u'post'])
    def add_data(self, request, pk=None):
        """Add data to collection."""
        collection = self.get_object()

        if 'ids' not in request.data:
            return Response({"error": "`ids`parameter is required"}, status=status.HTTP_400_BAD_REQUEST)

        missing = []
        for data_id in request.data['ids']:
            if not Data.objects.filter(pk=data_id).exists():
                missing.append(data_id)

        if missing:
            return Response(
                {"error": "Data objects with following ids are missing: {}".format(', '.join(missing))},
                status=status.HTTP_400_BAD_REQUEST)

        for data_id in request.data['ids']:
            collection.data.add(data_id)

        return Response()

    @detail_route(methods=[u'post'])
    def remove_data(self, request, pk=None):
        """Remove data from collection."""
        collection = self.get_object()

        if 'ids' not in request.data:
            return Response({"error": "`ids`parameter is required"}, status=status.HTTP_400_BAD_REQUEST)

        for data_id in request.data['ids']:
            collection.data.remove(data_id)

        return Response()



""".. Ignore pydocstyle D400.

=======================
Elastic Signal Handlers
=======================

"""
from __future__ import absolute_import, division, print_function, unicode_literals

from django.db.models.signals import post_delete, post_save
from django.dispatch import receiver

from guardian.models import GroupObjectPermission, UserObjectPermission

from .builder import index_builder


def _process_permission(perm):
    """Rebuild indexes affected by the given permission."""
    # XXX: Optimize: rebuild only permissions, not whole document
    codename = perm.permission.codename
    if not codename.startswith('view') and not codename.startswith('owner'):
        return

    index_builder.build(perm.content_object)


@receiver(post_save, sender=UserObjectPermission)
def add_user_permission(sender, instance, **kwargs):
    """Process indexes after adding user permission."""
    _process_permission(instance)


@receiver(post_save, sender=GroupObjectPermission)
def add_group_permission(sender, instance, **kwargs):
    """Process indexes after adding group permission."""
    _process_permission(instance)


@receiver(post_delete, sender=UserObjectPermission)
def remove_user_permission(sender, instance, **kwargs):
    """Process indexes after removing user permission."""
    _process_permission(instance)


@receiver(post_delete, sender=GroupObjectPermission)
def remove_group_permission(sender, instance, **kwargs):
    """Process indexes after removing group permission."""
    _process_permission(instance)



"""Workflow expression engine exceptions."""
from __future__ import absolute_import, division, print_function, unicode_literals


class EvaluationError(Exception):
    """Raised for errors that occur during expression evaluation."""

    pass



# -*- coding: utf-8 -*-

import os

# -- General configuration ------------------------------------------------

# The extension modules to enable.
extensions = ['resolwe.flow.utils.docs.autoprocess']

# Parent directory of all process definitions:
autoprocess_process_dir = os.path.normpath(os.path.abspath(os.path.dirname(__file__)) + '/processes')

# Base of the url to process source code:
autoprocess_source_base_url = 'https://github.com/genialis/resolwe-bio/blob/master/resolwe_bio/processes/'
